# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_service import service

from ovn_bgp_agent import config
from ovn_bgp_agent.drivers import driver_api


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class BGPAgent(service.Service):
    """BGP OVN Agent."""

    def __init__(self):
        super(BGPAgent, self).__init__()
        self.agent_driver = driver_api.AgentDriverBase.get_instance(
            CONF.driver)

    def start(self):
        LOG.info("Service '%s' starting", self.__class__.__name__)
        super(BGPAgent, self).start()
        self.agent_driver.start()

        LOG.info("Service '%s' started", self.__class__.__name__)
        sync_routes = loopingcall.FixedIntervalLoopingCall(self.sync)
        sync_routes.start(interval=CONF.reconcile_interval)
        sync_frr = loopingcall.FixedIntervalLoopingCall(self.frr_sync)
        sync_frr.start(interval=CONF.frr_reconcile_interval)

    def sync(self):
        LOG.info("Running reconciliation loop to ensure routes/rules are "
                 "in place.")
        try:
            self.agent_driver.sync()
        except Exception as e:
            LOG.exception("Unexpected exception while running the sync: %s", e)

    def frr_sync(self):
        LOG.info("Running reconciliation loop to ensure frr configuration is "
                 "in place.")
        try:
            self.agent_driver.frr_sync()
        except Exception as e:
            LOG.exception("Unexpected exception while running the frr sync: "
                          "%s", e)

    def wait(self):
        super(BGPAgent, self).wait()
        LOG.info("Service '%s' stopped", self.__class__.__name__)

    def stop(self, graceful=False):
        LOG.info("Service '%s' stopping", self.__class__.__name__)
        super(BGPAgent, self).stop(graceful)


def start():
    config.register_opts()
    config.init(sys.argv[1:])
    config.setup_logging()
    config.setup_privsep()

    bgp_agent_launcher = service.launch(config.CONF, BGPAgent())
    bgp_agent_launcher.wait()



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
from ovs.db import idl
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.schema.open_vswitch import impl_idl as idl_ovs
import socket
import tenacity

from ovn_bgp_agent import constants
from ovn_bgp_agent import exceptions as agent_exc
import ovn_bgp_agent.privileged.ovs_vsctl
from ovn_bgp_agent.utils import linux_net

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def _find_ovs_port(bridge):
    # TODO(ltomasbo): What happens if there are several patch ports on the
    # same bridge?
    ovs_port = None
    ovs_ports = ovn_bgp_agent.privileged.ovs_vsctl.ovs_vsctl(
        ['list-ports', bridge])[0].rstrip()
    for p in ovs_ports.split('\n'):
        if p.startswith(constants.OVS_PATCH_PROVNET_PORT_PREFIX):
            ovs_port = p
    return ovs_port


def get_bridge_flows(bridge, filter_=None):
    args = ['dump-flows', bridge]
    if filter_ is not None:
        args.append(filter_)
    return ovn_bgp_agent.privileged.ovs_vsctl.ovs_ofctl(
        args)[0].split('\n')[1:-1]


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(agent_exc.PortNotFound),
    wait=tenacity.wait_fixed(1),
    stop=tenacity.stop_after_delay(5),
    reraise=True)
def get_device_port_at_ovs(device):
    try:
        ofport = ovn_bgp_agent.privileged.ovs_vsctl.ovs_vsctl(
            ['get', 'Interface', device, 'ofport']
            )[0].rstrip()
    except Exception:
        raise agent_exc.PortNotFound(port=device)
    if ofport == '[]':
        # NOTE(ltomasbo): there is a chance the patch port interface was
        # created but not yet added to ovs bridge, therefore it exists but
        # has an empty ofport. We should retry in this case
        raise agent_exc.PortNotFound(port=device)
    return ofport


def get_ovs_ports_info(bridge):
    ovs_ports = ovn_bgp_agent.privileged.ovs_vsctl.ovs_vsctl(
        ['list-ports', bridge])[0].rstrip()
    return ovs_ports.split("\n")


def get_ovs_patch_ports_info(bridge, prefix='patch-provnet-'):
    in_ports = []
    ovs_ports = get_ovs_ports_info(bridge)
    for ovs_port in ovs_ports:
        if ovs_port.startswith(prefix):
            ovs_ofport = get_device_port_at_ovs(ovs_port)
            in_ports.append(ovs_ofport)
    return in_ports


def get_ovs_patch_port_ofport(patch):
    patch_name = "patch-{}-to-br-int".format(patch)
    return get_device_port_at_ovs(patch_name)


def ensure_mac_tweak_flows(bridge, mac, ports, cookie):
    cookie_id = "cookie={}/-1".format(cookie)
    current_flows = get_bridge_flows(bridge, cookie_id)
    flows_info = [flow.split("priority")[1].replace(" ", ",")
                  for flow in current_flows]

    for in_port in ports:
        exist_flow = False
        exist_flow_v6 = False
        flow = ("cookie={},priority=900,ip,in_port={},"
                "actions=mod_dl_dst:{},NORMAL".format(
                    cookie, in_port, mac))
        flow_v6 = ("cookie={},priority=900,ipv6,in_port={},"
                   "actions=mod_dl_dst:{},NORMAL".format(
                       cookie, in_port, mac))

        if flow.split("priority")[1] in flows_info:
            exist_flow = True
        if flow_v6.split("priority")[1] in flows_info:
            exist_flow_v6 = True

        if not exist_flow:
            ovn_bgp_agent.privileged.ovs_vsctl.ovs_ofctl(
                ['add-flow', bridge, flow])
        if not exist_flow_v6:
            ovn_bgp_agent.privileged.ovs_vsctl.ovs_ofctl(
                ['add-flow', bridge, flow_v6])


def remove_extra_ovs_flows(ovs_flows, bridge, cookie):
    expected_flows = []
    for port in ovs_flows[bridge].get('in_port'):
        pmm = ovs_flows[bridge].get('port-mac-mapping', {})
        lladdr = pmm.get(port, ovs_flows[bridge]['mac'])

        for flow in [
            # Add ipv4 flow in 'normal' and OpenFlow13 format
            "=900,ip,in_port={} actions=mod_dl_dst:{},NORMAL",
            "=900,ip,in_port={} actions=set_field:{}->eth_dst,NORMAL",

            # Add ipv6 flow in 'normal' and OpenFlow13 format
            "=900,ipv6,in_port={} actions=mod_dl_dst:{},NORMAL",
            "=900,ipv6,in_port={} actions=set_field:{}->eth_dst,NORMAL",
        ]:
            expected_flows.append(flow.format(port, lladdr))

    cookie_id = "cookie={}/-1".format(cookie)
    current_flows = get_bridge_flows(bridge, cookie_id)
    for flow in current_flows:
        if flow.split("priority")[1] not in expected_flows:
            del_flow = ('{},{}').format(
                cookie_id, flow.split("priority=900,")[1].split(" actions")[0])
            ovn_bgp_agent.privileged.ovs_vsctl.ovs_ofctl(
                ['del-flows', bridge, del_flow])


def ensure_flow(bridge, flow):
    ovn_bgp_agent.privileged.ovs_vsctl.ovs_ofctl(['add-flow', bridge, flow])


def ensure_evpn_ovs_flow(bridge, cookie, mac, output_port, port_dst, net,
                         strip_vlan=False):
    ovs_port = _find_ovs_port(bridge)
    if not ovs_port:
        return
    ovs_ofport = get_device_port_at_ovs(ovs_port)
    vrf_ofport = get_device_port_at_ovs(output_port)

    strip_vlan_opt = 'strip_vlan,' if strip_vlan else ''
    ip_version = linux_net.get_ip_version(net)
    port_dst_mac = linux_net.get_interface_address(port_dst)
    if ip_version == constants.IP_VERSION_6:
        flow = (
            "cookie={},priority=1000,ipv6,in_port={},dl_src:{},"
            "ipv6_src={} actions=mod_dl_dst:{},{}output={}".format(
                cookie, ovs_ofport, mac, net, port_dst_mac, strip_vlan_opt,
                vrf_ofport))
    else:
        flow = (
            "cookie={},priority=1000,ip,in_port={},dl_src:{},nw_src={}"
            "actions=mod_dl_dst:{},{}output={}".format(
                cookie, ovs_ofport, mac, net, port_dst_mac, strip_vlan_opt,
                vrf_ofport))
    ovn_bgp_agent.privileged.ovs_vsctl.ovs_ofctl(['add-flow', bridge, flow])


def remove_evpn_router_ovs_flows(bridge, cookie, mac):
    ovs_port = _find_ovs_port(bridge)
    if not ovs_port:
        return
    ovs_ofport = get_device_port_at_ovs(ovs_port)
    cookie_id = "cookie={}/-1".format(cookie)
    flow = ("{},ip,in_port={},dl_src:{}".format(
            cookie_id, ovs_ofport, mac))
    ovn_bgp_agent.privileged.ovs_vsctl.ovs_ofctl(['del-flows', bridge, flow])

    flow_v6 = ("{},ipv6,in_port={},dl_src:{}".format(cookie_id, ovs_ofport,
                                                     mac))
    ovn_bgp_agent.privileged.ovs_vsctl.ovs_ofctl(
        ['del-flows', bridge, flow_v6])


def remove_evpn_network_ovs_flow(bridge, cookie, mac, net):
    ovs_port = _find_ovs_port(bridge)
    if not ovs_port:
        return
    ovs_ofport = get_device_port_at_ovs(ovs_port)
    cookie_id = "cookie={}/-1".format(cookie)
    ip_version = linux_net.get_ip_version(net)
    if ip_version == constants.IP_VERSION_6:
        flow = ("{},ipv6,in_port={},dl_src:{},ipv6_src={}".format(
                cookie_id, ovs_ofport, mac, net))
    else:
        flow = ("{},ip,in_port={},dl_src:{},nw_src={}".format(
                cookie_id, ovs_ofport, mac, net))
    ovn_bgp_agent.privileged.ovs_vsctl.ovs_ofctl(['del-flows', bridge, flow])


def add_device_to_ovs_bridge(device, bridge, vlan_tag=None):
    args = ['--may-exist', 'add-port', bridge, device]
    if vlan_tag is not None:
        args.append('tag=%s' % vlan_tag)
    ovn_bgp_agent.privileged.ovs_vsctl.ovs_vsctl(args)


def del_device_from_ovs_bridge(device, bridge=None):
    args = ['--if-exists', 'del-port']
    if bridge:
        args.append(bridge)
    args.append(device)
    ovn_bgp_agent.privileged.ovs_vsctl.ovs_vsctl(args)


def add_vlan_port_to_ovs_bridge(bridge, vlan, vlan_tag):
    # ovs-vsctl add-port BRIDGE VLAN tag=VALN_ID
    # -- set interface VLAN type=internal
    args = [
        '--may-exist', 'add-port', bridge, vlan, 'tag={}'.format(vlan_tag),
        '--', 'set', 'interface', vlan, 'type=internal']
    ovn_bgp_agent.privileged.ovs_vsctl.ovs_vsctl(args)


def del_flow(flow, bridge, cookie):
    cookie_id = "cookie={}/-1".format(cookie)
    f = '{},priority{}'.format(
        cookie_id, flow.split(' actions')[0].split(' priority')[1])
    ovn_bgp_agent.privileged.ovs_vsctl.ovs_ofctl(
        ['--strict', 'del-flows', bridge, f])


def add_fdb_entry(bridge: str, port: str, mac: str, vlan: int = 0):
    ovn_bgp_agent.privileged.ovs_vsctl.ovs_appctl(
        ['fdb/add', bridge, port, str(vlan), mac]
    )


def get_flow_info(flow):
    # example:
    # cookie=0x3e7, duration=85.005s, table=0, n_packets=0,
    #  n_bytes=0, idle_age=65534, priority=1000,ip,in_port=1
    #  nw_src=20.0.0.0/24 actions=mod_dl_dst:1a:bd:c3:dc:6a:4c,
    # output:5
    flow_mac = flow_port = flow_nw_src = flow_ipv6_src = None
    try:
        flow_mac = flow.split('dl_src=')[1].split(',')[0]
        flow_port = flow.split('output:')[1].split(',')[0]
    except (IndexError, TypeError):
        pass
    flow_nw = flow.split('nw_src=')
    if len(flow_nw) == 2:
        flow_nw_src = flow_nw[1].split(' ')[0]
    flow_ipv6 = flow.split('ipv6_src=')
    if len(flow_ipv6) == 2:
        flow_ipv6_src = flow_ipv6[1].split(' ')[0]

    return {'mac': flow_mac, 'port': flow_port, 'nw_src': flow_nw_src,
            'ipv6_src': flow_ipv6_src}


class OvsIdl(object):
    def start(self, connection_string):
        helper = idlutils.get_schema_helper(connection_string,
                                            'Open_vSwitch')
        tables = ('Open_vSwitch', 'Bridge', 'Port', 'Interface')
        for table in tables:
            helper.register_table(table)
        ovs_idl = idl.Idl(connection_string, helper)
        ovs_idl._session.reconnect.set_probe_interval(60000)
        conn = connection.Connection(
            ovs_idl, timeout=180)
        self.idl_ovs = idl_ovs.OvsdbIdl(conn)

    def _get_from_ext_ids(self, key):
        return self.idl_ovs.db_get(
            'Open_vSwitch', '.', 'external_ids').execute()[key]

    def get_own_chassis_id(self):
        """Return the external_ids:system-id value of the Open_vSwitch table.

        """
        return self._get_from_ext_ids('system-id')

    def get_own_chassis_name(self):
        """Return the external_ids:hostname value of the Open_vSwitch table.

        If the value is not configured, it will fetch the hostname from the
        current machine.
        """
        try:
            return self._get_from_ext_ids('hostname')
        except KeyError:
            return socket.gethostname()

    def get_ovn_remote(self, nb=False):
        """Return the external_ids:ovn-remote value of the Open_vSwitch table.

        """
        if nb:
            return (CONF.ovn.ovn_nb_connection if CONF.ovn.ovn_nb_connection
                    else self._get_from_ext_ids('ovn-nb-remote'))
        return (CONF.ovn.ovn_sb_connection if CONF.ovn.ovn_sb_connection
                else self._get_from_ext_ids('ovn-remote'))

    def get_ovn_bridge_mappings(self, bridge=None):
        """Return a list of bridge mappings

        Return a list of bridge mappings based on the
        external_ids:ovn-bridge-mappings value of the Open_vSwitch table.
        """
        key = 'ovn-bridge-mappings'
        if bridge:
            key = key + '-' + str(bridge)
        try:
            return [i.strip() for i in
                    self._get_from_ext_ids(key).split(',')]
        except KeyError:
            return []



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shlex

from oslo_config import cfg
from oslo_log import log as logging
from oslo_privsep import priv_context

from ovn_bgp_agent import constants

LOG = logging.getLogger(__name__)

agent_opts = [
    cfg.IntOpt('reconcile_interval',
               help='Time (seconds) between re-sync actions.',
               default=300),
    cfg.IntOpt('frr_reconcile_interval',
               help='Time (seconds) between re-sync actions to ensure frr '
                    'configuration is correct, in case frr is restart.',
               default=15),
    cfg.BoolOpt('expose_tenant_networks',
                help='Expose VM IPs on tenant networks. '
                     'If this flag is enabled, it takes precedence over '
                     'expose_ipv6_gua_tenant_networks flag and all tenant '
                     'network IPs will be exposed.',
                default=False),
    cfg.StrOpt('advertisement_method_tenant_networks',
               help='The NB driver is capable of advertising the tenant '
                    'networks either per host or per subnet. '
                    'So either per /32 or /128 or per subnet like /24. '
                    'Choose "host" as value for this option to advertise per '
                    'host or choose "subnet" to announce per subnet prefix.',
               default=constants.ADVERTISEMENT_METHOD_HOST,
               choices=[constants.ADVERTISEMENT_METHOD_HOST,
                        constants.ADVERTISEMENT_METHOD_SUBNET]),
    cfg.BoolOpt('require_snat_disabled_for_tenant_networks',
                help='Require SNAT on the router port to be disabled before '
                     'exposing the tenant networks. Otherwise the exposed '
                     'tenant networks will be reachable from the outside, but'
                     'the connections set up from within the tenant vm will '
                     'always be SNAT-ed by the router, thus be the router ip. '
                     'When SNAT is disabled, OVN will do pure routing without '
                     'SNAT',
                default=False),
    cfg.BoolOpt('expose_ipv6_gua_tenant_networks',
                help='Expose only VM IPv6 IPs on tenant networks if they are '
                     'GUA. The expose_tenant_networks parameter takes '
                     'precedence over this one. So if it is set, all the '
                     'tenant network IPs will be exposed and not only the '
                     'IPv6 GUA IPs.',
                default=False),
    cfg.StrOpt('driver',
               help='Driver to be used',
               choices=('ovn_bgp_driver', 'ovn_evpn_driver',
                        'ovn_stretched_l2_bgp_driver', 'nb_ovn_bgp_driver'),
               default='ovn_bgp_driver'),
    cfg.StrOpt('ovsdb_connection',
               default='unix:/usr/local/var/run/openvswitch/db.sock',
               help='The connection string for the native OVSDB backend.\n'
                    'Use tcp:IP:PORT for TCP connection.\n'
                    'Use unix:FILE for unix domain socket connection.'),
    cfg.IntOpt('ovsdb_connection_timeout',
               default=180,
               help='Timeout in seconds for the OVSDB connection transaction'),
    cfg.StrOpt('bgp_AS',
               default='64999',
               help='AS number to be used by the Agent when running in BGP '
                    'mode and configuring the VRF route leaking.'),
    cfg.StrOpt('bgp_router_id',
               default=None,
               help='Router ID to be used by the Agent when running in BGP '
                    'mode and configuring the VRF route leaking.'),
    cfg.IPOpt('evpn_local_ip',
              default=None,
              help='IP address of local EVPN VXLAN (tunnel) endpoint. '
                   'This option can be used instead of the evpn_nic one. '
                   'If none specified, it will take the one from the '
                   'loopback device.'),
    cfg.StrOpt('evpn_nic',
               default=None,
               help='NIC with the IP address to use for the local EVPN '
                    'VXLAN (tunnel) endpoint. This option can be used '
                    'instead of the evpn_local_ip one. If none specified, '
                    'it will take the one from the loopback device.'),
    cfg.PortOpt('evpn_udp_dstport',
                default=4789,
                help='The UDP port used for EVPN VXLAN communication. By '
                     'default 4789 is being used.'),
    cfg.BoolOpt('clear_vrf_routes_on_startup',
                help='If enabled, all routes are removed from the VRF table'
                     '(specified by bgp_vrf_table_id option) at startup.',
                default=False),
    cfg.BoolOpt('delete_vrf_on_disconnect',
                help='If enabled agent will take care of completely deleting'
                     'VRF from both kernel and FRR configuration.'
                     'Disabling option will keep VRF even when agent considers'
                     'its presence as redundant',
                default=True),
    cfg.StrOpt('bgp_nic',
               default='bgp-nic',
               help='The name of the interface used within the VRF '
                    '(bgp_vrf option) to expose the IPs and/or Networks.'),
    cfg.StrOpt('bgp_vrf',
               default='bgp-vrf',
               help='The name of the VRF to be used to expose the IPs '
                    'and/or Networks through BGP.'),
    cfg.IntOpt('bgp_vrf_table_id',
               default=10,
               help='The Routing Table ID that the VRF (bgp_vrf option) '
                    'should use. If it does not exist, this table will be '
                    'created.'),
    cfg.ListOpt('address_scopes',
                default=None,
                help='Allows to filter on the address scope. Only networks'
                     ' with the same address scope on the provider and'
                     ' internal interface are announced.'),
    cfg.StrOpt('exposing_method',
               default='underlay',
               choices=('underlay', 'l2vni', 'vrf', 'dynamic', 'ovn'),
               help='The exposing mechanism to be used. underlay is the '
                    'default and simply expose it on the default/plain '
                    'network.'
                    'With l2vni the l2 is extended over the l3 infrastructure '
                    'and it is exposed on a given VNI (Type-2).'
                    'With vrf the routes are exposed in different VRFs/VNIs '
                    '(Type-5).'
                    'With dynamic, a mix between underlay, l2vni and vrf is '
                    'used, depending on the information annotated on the '
                    'ports. '
                    'Finally, with ovn, instead of using kernel networking a '
                    'dedicated ovn cluster per node is used for the traffic '
                    'redirection'),
]

root_helper_opts = [
    cfg.StrOpt('root_helper', default='sudo',
               help=("Root helper application. "
                     "Use 'sudo ovn-bgp-agent-rootwrap  "
                     "/etc/ovn-bgp-agent/rootwrap.conf' to use the real "
                     "root filter facility. Change to 'sudo' to skip the "
                     "filtering and just run the command directly.")),
    cfg.BoolOpt('use_helper_for_ns_read',
                default=True,
                help=("Use the root helper when listing the namespaces on a "
                      "system. This may not be required depending on the "
                      "security configuration. If the root helper is "
                      "not required, set this to False for a performance "
                      "improvement.")),
    # We can't just use root_helper=sudo ovn-bgp-agent-rootwrap-daemon $cfg
    # because it isn't appropriate for long-lived processes spawned with
    # create_process. Having a bool use_rootwrap_daemon option precludes
    # specifying the rootwrap daemon command, which may be necessary for Xen?
    cfg.StrOpt('root_helper_daemon',
               help=("""
Root helper daemon application to use when possible.

Use 'sudo ovn-bgp-agent-rootwrap-daemon /etc/ovn-bgp-agent/rootwrap.conf'
to run rootwrap in "daemon mode" which has been reported to improve
performance at scale. For more information on running rootwrap in
"daemon mode", see:

https://docs.openstack.org/oslo.rootwrap/latest/user/usage.html#daemon-mode
""")),
]

ovn_opts = [
    cfg.StrOpt('ovn_sb_private_key',
               default='/etc/pki/tls/private/ovn_bgp_agent.key',
               deprecated_group='DEFAULT',
               help='The PEM file with private key for SSL connection to '
                    'OVN-SB-DB'),
    cfg.StrOpt('ovn_sb_certificate',
               default='/etc/pki/tls/certs/ovn_bgp_agent.crt',
               deprecated_group='DEFAULT',
               help='The PEM file with certificate that certifies the '
                    'private key specified in ovn_sb_private_key'),
    cfg.StrOpt('ovn_sb_ca_cert',
               default='/etc/ipa/ca.crt',
               deprecated_group='DEFAULT',
               help='The PEM file with CA certificate that OVN should use to'
                    ' verify certificates presented to it by SSL peers'),
    cfg.StrOpt('ovn_sb_connection',
               default='',
               deprecated_group='DEFAULT',
               help='The connection string for the OVN_Southbound OVSDB.\n'
                    'Use tcp:IP:PORT for TCP connection.\n'
                    'Use unix:FILE for unix domain socket connection.'),
    cfg.StrOpt('ovn_nb_private_key',
               default='/etc/pki/tls/private/ovn_bgp_agent.key',
               deprecated_group='DEFAULT',
               help='The PEM file with private key for SSL connection to '
                    'OVN-NB-DB'),
    cfg.StrOpt('ovn_nb_certificate',
               default='/etc/pki/tls/certs/ovn_bgp_agent.crt',
               deprecated_group='DEFAULT',
               help='The PEM file with certificate that certifies the '
                    'private key specified in ovn_nb_private_key'),
    cfg.StrOpt('ovn_nb_ca_cert',
               default='/etc/ipa/ca.crt',
               deprecated_group='DEFAULT',
               help='The PEM file with CA certificate that OVN should use to'
                    ' verify certificates presented to it by SSL peers'),
    cfg.StrOpt('ovn_nb_connection',
               default='',
               deprecated_group='DEFAULT',
               help='The connection string for the OVN_Northbound OVSDB.\n'
                    'Use tcp:IP:PORT for TCP connection.\n'
                    'Use unix:FILE for unix domain socket connection.'),
]

local_ovn_cluster_opts = [
    cfg.StrOpt('ovn_nb_connection',
               default='unix:/var/run/ovn/ovnnb_db.sock',
               help='The connection string for the OVN_Northbound OVSDB.\n'
                    'Use tcp:IP:PORT for TCP connection.\n'
                    'Use unix:FILE for unix domain socket connection.'),
    cfg.ListOpt('external_nics',
                default=[],
                help='List of NICS that the local OVN cluster needs to be '
                     'connected to for the external connectivity.'),
    cfg.ListOpt('peer_ips',
                default=[],
                help='List of peer IPs used for redirecting the outgoing '
                     'traffic (ECMP supported).'),
    cfg.ListOpt('provider_networks_pool_prefixes',
                default=['192.168.0.0/16'],
                help='List of prefixes for provider networks'),
    cfg.StrOpt('bgp_chassis_id',
               default='bgp',
               help='The chassis_id used for the ovn-controller instance'
                    ' related to the node-local OVN instance. Used as a'
                    ' suffix for getting instance-specific options'
                    ' from OVSDB. This option has effect only when the OVN'
                    ' NB driver is used.'),
]

CONF = cfg.CONF
EXTRA_LOG_LEVEL_DEFAULTS = [
    'oslo.privsep.daemon=INFO'
]

logging.register_options(CONF)


def register_opts():
    CONF.register_opts(agent_opts)
    CONF.register_opts(root_helper_opts, "agent")
    CONF.register_opts(ovn_opts, "ovn")
    CONF.register_opts(local_ovn_cluster_opts, "local_ovn_cluster")


def init(args, **kwargs):
    CONF(args=args, project='bgp-agent', **kwargs)


def setup_logging():
    logging.set_defaults(default_log_levels=logging.get_default_log_levels() +
                         EXTRA_LOG_LEVEL_DEFAULTS)
    logging.setup(CONF, 'bgp-agent')
    LOG.info("Logging enabled!")


def get_root_helper(conf):
    return conf.agent.root_helper


def setup_privsep():
    priv_context.init(root_helper=shlex.split(get_root_helper(cfg.CONF)))


def list_opts():
    return [
        ("DEFAULT", agent_opts),
        ("agent", root_helper_opts),
        ("ovn", ovn_opts),
        ("local_ovn_cluster", local_ovn_cluster_opts),
    ]



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_concurrency import processutils
from oslo_log import log as logging

import ovn_bgp_agent.privileged.ovs_vsctl

LOG = logging.getLogger(__name__)


@ovn_bgp_agent.privileged.ovs_vsctl_cmd.entrypoint
def ovs_cmd(command, args, timeout=None):
    full_args = [command]
    if timeout is not None:
        full_args += ['--timeout=%s' % timeout]
    full_args += args
    try:
        return processutils.execute(*full_args)
    except Exception:
        LOG.error("Unable to execute %s %s", command, full_args)
        raise


def ovs_vsctl(args, timeout=None):
    return ovs_cmd('ovs-vsctl', args, timeout)


def ovs_ofctl(args, timeout=None):
    try:
        return ovs_cmd('ovs-ofctl', args, timeout)
    except processutils.ProcessExecutionError:
        return ovs_cmd('ovs-ofctl', args + ['-O', 'OpenFlow13'], timeout)


def ovs_appctl(args):
    return ovs_cmd('ovs-appctl', args)



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack import ovn_bgp_driver
from ovn_bgp_agent.drivers.openstack.utils import bgp as bgp_utils
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent.drivers.openstack.utils import frr
from ovn_bgp_agent.drivers.openstack.utils import ovn
from ovn_bgp_agent.drivers.openstack.utils import ovs
from ovn_bgp_agent.drivers.openstack.utils import wire as wire_utils
from ovn_bgp_agent import exceptions as agent_exc
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests.unit import fakes
from ovn_bgp_agent.utils import linux_net

CONF = cfg.CONF


class TestOVNBGPDriver(test_base.TestCase):

    def setUp(self):
        super(TestOVNBGPDriver, self).setUp()
        CONF.set_override('expose_tenant_networks', True)
        self.bridge = 'fake-bridge'
        self.bgp_driver = ovn_bgp_driver.OVNBGPDriver()
        self.bgp_driver._post_fork_event = mock.Mock()
        self.bgp_driver.sb_idl = mock.Mock()
        self.sb_idl = self.bgp_driver.sb_idl
        self.bgp_driver.chassis = 'fake-chassis'
        self.bgp_driver.ovn_routing_tables = {self.bridge: 'fake-table'}
        self.bgp_driver.ovn_bridge_mappings = {'fake-network': self.bridge}

        self.mock_sbdb = mock.patch.object(ovn, 'OvnSbIdl').start()
        self.mock_ovs_idl = mock.patch.object(ovs, 'OvsIdl').start()
        self.ipv4 = '192.168.1.17'
        self.ipv6 = '2002::1234:abcd:ffff:c0a8:101'
        self.fip = '172.24.4.33'
        self.mac = 'aa:bb:cc:dd:ee:ff'
        self.loadbalancer_vip_port = 'fake-lb-vip-port'
        self.bgp_driver.ovs_idl = self.mock_ovs_idl

        self.cr_lrp0 = 'cr-fake-logical-port'
        self.cr_lrp1 = 'cr-fake-logical-port1'
        self.lrp0 = 'lrp-fake-logical-port'
        self.bgp_driver.ovn_local_cr_lrps = {
            self.cr_lrp0: {'provider_datapath': 'fake-provider-dp',
                           'router_datapath': 'fake-router-dp',
                           'ips': [self.fip],
                           'subnets_datapath': {self.lrp0: 'fake-lrp-dp'},
                           'subnets_cidr': ['192.168.1.1/24'],
                           'provider_ovn_lbs': [],
                           'bridge_device': self.bridge,
                           'bridge_vlan': None},
            self.cr_lrp1: {'provider_datapath': 'fake-provider-dp2'}}
        self.bgp_driver.provider_ovn_lbs = {
            self.loadbalancer_vip_port: {'ips': [self.ipv4, self.ipv6],
                                         'gateway_port': self.cr_lrp0}}

    @mock.patch.object(linux_net, 'ensure_ovn_device')
    @mock.patch.object(linux_net, 'ensure_vrf')
    @mock.patch.object(frr, 'vrf_leak')
    def test_start(self, mock_vrf, *args):
        self.bgp_driver.start()

        mock_vrf.assert_called_once_with(
            CONF.bgp_vrf, CONF.bgp_AS, CONF.bgp_router_id,
            template=frr.LEAK_VRF_TEMPLATE)
        # Assert connections were started
        self.mock_ovs_idl().start.assert_called_once_with(
            CONF.ovsdb_connection)
        self.mock_sbdb().start.assert_called_once_with()

    @mock.patch.object(linux_net, 'ensure_ovn_device')
    @mock.patch.object(frr, 'vrf_leak')
    @mock.patch.object(linux_net, 'ensure_vrf')
    def test_frr_sync(self, mock_ensure_vrf, mock_vrf_leak,
                      mock_ensure_ovn_dev):
        self.bgp_driver.frr_sync()

        mock_ensure_vrf.assert_called_once_with(
            CONF.bgp_vrf, CONF.bgp_vrf_table_id)
        mock_vrf_leak.assert_called_once_with(
            CONF.bgp_vrf, CONF.bgp_AS, CONF.bgp_router_id,
            template=frr.LEAK_VRF_TEMPLATE)
        mock_ensure_ovn_dev.assert_called_once_with(
            CONF.bgp_nic, CONF.bgp_vrf)

    @mock.patch.object(wire_utils, 'delete_vlan_devices_leftovers')
    @mock.patch.object(linux_net, 'delete_bridge_ip_routes')
    @mock.patch.object(linux_net, 'delete_ip_rules')
    @mock.patch.object(linux_net, 'delete_exposed_ips')
    @mock.patch.object(ovs, 'remove_extra_ovs_flows')
    @mock.patch.object(ovs, 'ensure_mac_tweak_flows')
    @mock.patch.object(ovs, 'get_ovs_patch_ports_info')
    @mock.patch.object(linux_net, 'get_ovn_ip_rules')
    @mock.patch.object(linux_net, 'get_exposed_ips')
    @mock.patch.object(linux_net, 'get_interface_address')
    @mock.patch.object(linux_net, 'ensure_vlan_device_for_network')
    @mock.patch.object(linux_net, 'ensure_routing_table_for_bridge')
    @mock.patch.object(linux_net, 'ensure_arp_ndp_enabled_for_bridge')
    def test_sync(
            self, mock_ensure_arp, mock_routing_bridge,
            mock_ensure_vlan_network, mock_nic_address, mock_exposed_ips,
            mock_get_ip_rules, mock_get_patch_ports, mock_ensure_mac,
            mock_remove_flows, mock_del_exposed_ips, mock_del_ip_rules,
            mock_del_ip_routes, mock_vlan_leftovers):
        self.mock_ovs_idl.get_ovn_bridge_mappings.return_value = [
            'net0:bridge0', 'net1:bridge1']
        self.sb_idl.get_network_vlan_tag_by_network_name.side_effect = (
            [10], [11])
        fake_ip_rules = 'fake-ip-rules'
        mock_get_ip_rules.return_value = fake_ip_rules
        ips = [self.ipv4, self.ipv6]
        mock_exposed_ips.return_value = ips
        self.sb_idl.get_ports_on_chassis.return_value = [
            'fake-port0', 'fake-port1']
        self.sb_idl.get_cr_lrp_ports_on_chassis.return_value = [
            'fake-cr-port0', 'fake-cr-port1']

        mock_ensure_port_exposed = mock.patch.object(
            self.bgp_driver, '_ensure_port_exposed').start()
        mock_ensure_cr_port_exposed = mock.patch.object(
            self.bgp_driver, '_ensure_cr_lrp_associated_ports_exposed').start()
        mock_routing_bridge.return_value = ['fake-route']
        mock_nic_address.return_value = self.mac
        mock_get_patch_ports.return_value = [1, 2]

        self.bgp_driver.sync()

        expected_calls = [mock.call('bridge0', 1, [10]),
                          mock.call('bridge1', 2, [11])]
        mock_ensure_arp.assert_has_calls(expected_calls)

        expected_calls = [mock.call({}, 'bridge0', CONF.bgp_vrf_table_id),
                          mock.call({}, 'bridge1', CONF.bgp_vrf_table_id)]
        mock_routing_bridge.assert_has_calls(expected_calls)

        expected_calls = [mock.call('bridge0', 10), mock.call('bridge1', 11)]
        mock_ensure_vlan_network.assert_has_calls(expected_calls)

        expected_calls = [
            mock.call('bridge0'), mock.call('bridge1')]
        mock_get_patch_ports.assert_has_calls(expected_calls)
        expected_calls = [
            mock.call('bridge0', mock.ANY, [1, 2], constants.OVS_RULE_COOKIE),
            mock.call('bridge1', mock.ANY, [1, 2], constants.OVS_RULE_COOKIE)]
        mock_ensure_mac.assert_has_calls(expected_calls)

        expected_calls = [
            mock.call(mock.ANY, 'bridge0', constants.OVS_RULE_COOKIE),
            mock.call(mock.ANY, 'bridge1', constants.OVS_RULE_COOKIE)]
        mock_remove_flows.assert_has_calls(expected_calls)

        expected_calls = [mock.call('fake-port0', ips, fake_ip_rules),
                          mock.call('fake-port1', ips, fake_ip_rules)]
        mock_ensure_port_exposed.assert_has_calls(expected_calls)

        expected_calls = [mock.call('fake-cr-port0', ips, fake_ip_rules),
                          mock.call('fake-cr-port1', ips, fake_ip_rules)]
        mock_ensure_cr_port_exposed.assert_has_calls(expected_calls)

        mock_del_exposed_ips.assert_called_once_with(
            ips, CONF.bgp_nic)
        mock_del_ip_rules.assert_called_once_with(fake_ip_rules)
        mock_del_ip_routes.assert_called_once_with(
            {}, mock.ANY,
            {'bridge0': ['fake-route'], 'bridge1': ['fake-route']})

        mock_get_ip_rules.assert_called_once_with(mock.ANY)
        mock_vlan_leftovers.assert_called_once_with(
            self.sb_idl, self.bgp_driver.ovn_bridge_mappings)

    @mock.patch.object(linux_net, 'get_ip_version')
    def test__ensure_cr_lrp_associated_ports_exposed(self, mock_ip_version):
        mock_expose_ip = mock.patch.object(
            self.bgp_driver, '_expose_ip').start()
        mock_ip_version.side_effect = (constants.IP_VERSION_4,
                                       constants.IP_VERSION_6)
        patch_port_row = fakes.create_object({'name': 'patch-port'})
        self.sb_idl.get_cr_lrp_nat_addresses_info.return_value = (
            [self.ipv4, self.ipv6], patch_port_row)

        exposed_ips = [self.ipv4, '192.168.1.20']
        mock_expose_ip.return_value = exposed_ips
        ip_rules = {"{}/128".format(self.ipv6): 'fake-rules'}
        self.bgp_driver._ensure_cr_lrp_associated_ports_exposed(
            'fake-cr-lrp', exposed_ips, ip_rules)

        mock_expose_ip.assert_called_once_with(
            [self.ipv4, self.ipv6], patch_port_row,
            associated_port='fake-cr-lrp')
        self.assertEqual(['192.168.1.20'], exposed_ips)

    def test__ensure_port_exposed(self):
        mock_expose_ip = mock.patch.object(
            self.bgp_driver, '_expose_ip').start()
        mock_expose_ip.return_value = [self.ipv4, self.ipv6]
        port = fakes.create_object({
            'name': 'fake-port',
            'type': '',
            'mac': ['{} {} {}'.format(self.mac, self.ipv4, self.ipv6)]})

        exposed_ips = [self.ipv4, self.ipv6]
        ip_rules = {"{}/128".format(self.ipv6): 'fake-rules'}
        self.bgp_driver._ensure_port_exposed(port, exposed_ips, ip_rules)

        mock_expose_ip.assert_called_once_with(
            [self.ipv4, self.ipv6], port)
        self.assertEqual([], exposed_ips)
        self.assertEqual({}, ip_rules)

    def test__ensure_port_exposed_fip(self):
        fip = '172.24.4.225'
        mock_expose_ip = mock.patch.object(
            self.bgp_driver, '_expose_ip').start()
        mock_expose_ip.return_value = [fip]
        port = fakes.create_object({
            'name': 'fake-port',
            'type': '',
            'mac': ['{} {} {}'.format(self.mac, self.ipv4, self.ipv6)]})

        exposed_ips = [self.ipv4, fip]
        ip_rules = {"{}/128".format(self.ipv6): 'fake-rules'}
        self.bgp_driver._ensure_port_exposed(port, exposed_ips, ip_rules)

        mock_expose_ip.assert_called_once_with(
            [self.ipv4, self.ipv6], port)
        self.assertEqual([self.ipv4], exposed_ips)
        self.assertEqual({"{}/128".format(self.ipv6): 'fake-rules'}, ip_rules)

    def test__ensure_port_exposed_fip_unknown_mac(self):
        fip = '172.24.4.225'
        mock_expose_ip = mock.patch.object(
            self.bgp_driver, '_expose_ip').start()
        mock_expose_ip.return_value = [fip]
        port = fakes.create_object({
            'name': 'fake-port',
            'type': '',
            'mac': ['unknown'],
            'datapath': 'fake-dp'})

        exposed_ips = [self.ipv4, fip]
        ip_rules = {"{}/128".format(self.ipv6): 'fake-rules'}
        self.sb_idl.is_provider_network.return_value = False

        self.bgp_driver._ensure_port_exposed(port, exposed_ips, ip_rules)

        mock_expose_ip.assert_called_once_with([], port)
        self.assertEqual([self.ipv4], exposed_ips)
        self.assertEqual({"{}/128".format(self.ipv6): 'fake-rules'}, ip_rules)

    def test__ensure_port_exposed_wrong_port_type(self):
        mock_expose_ip = mock.patch.object(
            self.bgp_driver, '_expose_ip').start()
        port = fakes.create_object({
            'name': 'fake-port',
            'type': 'non-existing-type',
            'mac': ['{} {} {}'.format(self.mac, self.ipv4, self.ipv6)]})

        self.bgp_driver._ensure_port_exposed(port, [], {})

        # Assert it was never called, the method should just return if
        # the port type is not OVN_VIF_PORT_TYPES
        mock_expose_ip.assert_not_called()

    @mock.patch.object(wire_utils, '_ensure_updated_mac_tweak_flows')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    def test__expose_provider_port(self, mock_add_rule, mock_add_route,
                                   mock_add_ips_dev, mock_ensure_mac_tweak):
        port_ips = [self.ipv4]
        provider_datapath = 'fake-provider-dp'
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        self.bgp_driver._expose_provider_port(port_ips, provider_datapath)

        mock_ensure_mac_tweak.assert_called_once_with(mock.ANY, self.bridge,
                                                      {})
        mock_add_ips_dev.assert_called_once_with(
            CONF.bgp_nic, [self.ipv4])
        mock_add_rule.assert_called_once_with(
            self.ipv4, 'fake-table', dev='fake-bridge')
        mock_add_route.assert_called_once_with(
            mock.ANY, self.ipv4, 'fake-table', self.bridge, vlan=10)

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    def test__expose_provider_port_no_device(self, mock_add_rule,
                                             mock_add_route, mock_add_ips_dev):
        port_ips = [self.ipv4]
        provider_datapath = 'fake-provider-dp'
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (None, None)
        ret = self.bgp_driver._expose_provider_port(port_ips,
                                                    provider_datapath)

        self.assertEqual(False, ret)
        mock_add_ips_dev.assert_not_called()
        mock_add_rule.assert_not_called()
        mock_add_route.assert_not_called()

    @mock.patch.object(wire_utils, '_ensure_updated_mac_tweak_flows')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    def test__expose_provider_port_invalid_ip(
            self, mock_add_rule, mock_add_route, mock_add_ips_dev,
            mock_ensure_mac_tweak):
        port_ips = [self.ipv4]
        provider_datapath = 'fake-provider-dp'
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        mock_add_rule.side_effect = agent_exc.InvalidPortIP(ip=self.ipv4)
        self.sb_idl.get_localnet_for_datapath.return_value = 'fake-localnet'
        ret = self.bgp_driver._expose_provider_port(port_ips,
                                                    provider_datapath)

        self.assertEqual(False, ret)
        mock_add_ips_dev.assert_not_called()
        mock_add_rule.assert_called_once_with(
            self.ipv4, 'fake-table', dev='fake-bridge')
        mock_add_route.assert_not_called()
        mock_ensure_mac_tweak.assert_not_called()

    @mock.patch.object(wire_utils, '_ensure_updated_mac_tweak_flows')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    def test__expose_provider_port_with_lladdr(
            self, mock_add_rule, mock_add_route, mock_add_ips_dev,
            mock_ensure_mac_tweak):
        port_ips = [self.ipv4]
        provider_datapath = 'fake-provider-dp'
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        self.bgp_driver._expose_provider_port(port_ips, provider_datapath,
                                              lladdr='fake-mac')
        mock_ensure_mac_tweak.assert_called_once_with(mock.ANY, self.bridge,
                                                      {})
        mock_add_ips_dev.assert_called_once_with(
            CONF.bgp_nic, [self.ipv4])
        mock_add_rule.assert_called_once_with(
            self.ipv4, 'fake-table', dev='{}.{}'.format(self.bridge, 10),
            lladdr='fake-mac')
        mock_add_route.assert_called_once_with(
            mock.ANY, self.ipv4, 'fake-table', self.bridge, vlan=10)

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__expose_tenant_port(self, mock_ip_version, mock_add_ips_dev):
        tenant_port = fakes.create_object({
            'name': 'fake-port',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.10 192.168.1.11'],
            'chassis': 'fake-chassis1',
            'external_ids': {}})
        ip_version = constants.IP_VERSION_4

        mock_ip_version.return_value = constants.IP_VERSION_4

        self.bgp_driver._expose_tenant_port(tenant_port, ip_version)

        expected_calls = [mock.call(CONF.bgp_nic, ['192.168.1.10']),
                          mock.call(CONF.bgp_nic, ['192.168.1.11'])]
        mock_add_ips_dev.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__expose_tenant_port_ovn_lb(self, mock_ip_version,
                                        mock_add_ips_dev):
        tenant_port = fakes.create_object({
            'name': 'fake-port',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': '',
            'external_ids': {'neutron:cidrs': '192.168.1.10/24'},
            'up': [False]})
        ip_version = constants.IP_VERSION_4

        mock_ip_version.return_value = constants.IP_VERSION_4

        self.bgp_driver._expose_tenant_port(tenant_port, ip_version)

        mock_add_ips_dev.assert_called_once_with(CONF.bgp_nic,
                                                 ['192.168.1.10'])

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__expose_tenant_port_no_ip(self, mock_ip_version,
                                       mock_add_ips_dev):
        tenant_port = fakes.create_object({
            'name': 'fake-port',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': ['aa:bb:cc:dd:ee:ee'],
            'chassis': 'fake-chassis1',
            'external_ids': {}})
        ip_version = constants.IP_VERSION_4

        mock_ip_version.return_value = constants.IP_VERSION_4

        self.bgp_driver._expose_tenant_port(tenant_port, ip_version)

        mock_ip_version.assert_not_called()
        mock_add_ips_dev.assert_not_called()

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__expose_tenant_port_no_mac(self, mock_ip_version,
                                        mock_add_ips_dev):
        tenant_port = fakes.create_object({
            'name': 'fake-port',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': 'fake-chassis1',
            'external_ids': {}})
        ip_version = constants.IP_VERSION_4

        mock_ip_version.return_value = constants.IP_VERSION_4

        self.bgp_driver._expose_tenant_port(tenant_port, ip_version)

        mock_ip_version.assert_not_called()
        mock_add_ips_dev.assert_not_called()

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__expose_tenant_port_unknown_mac(self, mock_ip_version,
                                             mock_add_ips_dev):
        tenant_port = fakes.create_object({
            'name': 'fake-port',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': ['unknown'],
            'chassis': 'fake-chassis1',
            'external_ids': {'neutron:cidrs': '192.168.1.10/24'},
            'up': [False]})
        ip_version = constants.IP_VERSION_4

        mock_ip_version.return_value = constants.IP_VERSION_4

        self.bgp_driver._expose_tenant_port(tenant_port, ip_version)

        mock_add_ips_dev.assert_called_once_with(CONF.bgp_nic,
                                                 ['192.168.1.10'])

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__expose_tenant_port_wrong_type(self, mock_ip_version,
                                            mock_add_ips_dev):
        tenant_port = fakes.create_object({
            'name': 'fake-port',
            'type': constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.10 192.168.1.11'],
            'chassis': 'fake-chassis1',
            'external_ids': {}})
        ip_version = constants.IP_VERSION_4

        mock_ip_version.return_value = constants.IP_VERSION_4

        self.bgp_driver._expose_tenant_port(tenant_port, ip_version)

        mock_ip_version.assert_not_called()
        mock_add_ips_dev.assert_not_called()

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__expose_tenant_port_no_chassis(self, mock_ip_version,
                                            mock_add_ips_dev):
        tenant_port = fakes.create_object({
            'name': 'fake-port',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.10 192.168.1.11'],
            'chassis': '',
            'external_ids': {}})
        ip_version = constants.IP_VERSION_4

        mock_ip_version.return_value = constants.IP_VERSION_4

        self.bgp_driver._expose_tenant_port(tenant_port, ip_version)

        mock_ip_version.assert_not_called()
        mock_add_ips_dev.assert_not_called()

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    def test__withdraw_provider_port(self, mock_del_rule, mock_del_route,
                                     mock_del_ips_dev):
        port_ips = [self.ipv4]
        provider_datapath = 'fake-provider-dp'
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        self.bgp_driver._withdraw_provider_port(port_ips, provider_datapath)

        mock_del_ips_dev.assert_called_once_with(
            CONF.bgp_nic, [self.ipv4])
        mock_del_rule.assert_called_once_with(
            self.ipv4, 'fake-table', dev='fake-bridge')
        mock_del_route.assert_called_once_with(
            mock.ANY, self.ipv4, 'fake-table', self.bridge, vlan=10)

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    def test__withdraw_provider_port_no_device(self, mock_del_rule,
                                               mock_del_route,
                                               mock_del_ips_dev):
        port_ips = [self.ipv4]
        provider_datapath = 'fake-provider-dp'
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (None, None)
        ret = self.bgp_driver._withdraw_provider_port(port_ips,
                                                      provider_datapath)

        self.assertEqual(False, ret)
        mock_del_ips_dev.assert_called_once_with(
            CONF.bgp_nic, [self.ipv4])
        mock_del_rule.assert_not_called()
        mock_del_route.assert_not_called()

    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'del_ips_from_dev')
    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    def test__withdraw_provider_port_lladdr(
            self, mock_del_rule, mock_del_route, mock_del_ips_dev,
            mock_ip_version):
        port_ips = [self.ipv4]
        provider_datapath = 'fake-provider-dp'
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        dev = '{}.{}'.format(self.bridge, 10)
        mock_ip_version.return_value = constants.IP_VERSION_4
        self.bgp_driver._withdraw_provider_port(port_ips, provider_datapath,
                                                lladdr='fake-mac')

        mock_del_ips_dev.assert_called_once_with(
            CONF.bgp_nic, [self.ipv4])
        mock_del_rule.assert_called_once_with(
            '{}/32'.format(self.ipv4), 'fake-table', dev=dev,
            lladdr='fake-mac')
        mock_del_route.assert_called_once_with(
            mock.ANY, self.ipv4, 'fake-table', self.bridge, vlan=10)

    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'del_ips_from_dev')
    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    def test__withdraw_provider_port_lladdr_ipv6(
            self, mock_del_rule, mock_del_route, mock_del_ips_dev,
            mock_ip_version):
        port_ips = [self.ipv6]
        provider_datapath = 'fake-provider-dp'
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        dev = '{}.{}'.format(self.bridge, 10)
        mock_ip_version.return_value = constants.IP_VERSION_6
        self.bgp_driver._withdraw_provider_port(port_ips, provider_datapath,
                                                lladdr='fake-mac')

        mock_del_ips_dev.assert_called_once_with(
            CONF.bgp_nic, [self.ipv6])
        mock_del_rule.assert_called_once_with(
            '{}/128'.format(self.ipv6), 'fake-table', dev=dev,
            lladdr='fake-mac')
        mock_del_route.assert_called_once_with(
            mock.ANY, self.ipv6, 'fake-table', self.bridge, vlan=10)

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__process_lrp_port(self, mock_ip_version, mock_add_rule,
                               mock_add_route, mock_add_ips_dev):
        mock_ip_version.return_value = constants.IP_VERSION_4
        gateway = {}
        gateway['ips'] = ['{}/32'.format(self.fip),
                          '2003::1234:abcd:ffff:c0a8:102/128']
        gateway['provider_datapath'] = 'bc6780f4-9510-4270-b4d2-b8d5c6802713'
        gateway['subnets_datapath'] = {}
        gateway['subnets_cidr'] = []
        gateway['bridge_device'] = self.bridge
        gateway['bridge_vlan'] = 10
        self.bgp_driver.ovn_local_cr_lrps = {'gateway_port': gateway}

        router_port = fakes.create_object({
            'chassis': [],
            'mac': ['{} {}/32'.format(self.mac, self.ipv4)],
            'logical_port': 'lrp-fake-logical-port',
            'options': {'peer': 'fake-peer'}})

        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)

        dp_port0 = fakes.create_object({
            'name': 'fake-port-dp0',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.10 192.168.1.11'],
            'chassis': 'fake-chassis1',
            'external_ids': {}})
        dp_port1 = fakes.create_object({
            'name': 'fake-port-dp1',
            'type': 'fake-type',
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.12 192.168.1.13'],
            'chassis': 'fake-chassis2',
            'external_ids': {}})
        dp_port2 = fakes.create_object({
            'name': 'fake-port-dp2',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': 'fake-chassis2',
            'external_ids': {}})
        dp_port3 = fakes.create_object({
            'name': 'fake-port-dp3',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': '',
            'up': [False],
            'external_ids': {}})
        dp_port4 = fakes.create_object({
            'name': 'fake-port-dp4',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': '',
            'up': [False],
            'external_ids': {
                constants.OVN_CIDRS_EXT_ID_KEY: "192.168.1.13/24"}})
        self.sb_idl.get_ports_on_datapath.return_value = [dp_port0, dp_port1,
                                                          dp_port2, dp_port3,
                                                          dp_port4]

        self.bgp_driver._process_lrp_port(router_port, 'gateway_port')

        # Assert that the add methods were called
        mock_add_rule.assert_called_once_with(
            '{}/32'.format(self.ipv4), 'fake-table')
        mock_add_route.assert_called_once_with(
            mock.ANY, self.ipv4, 'fake-table', self.bridge,
            vlan=10, mask='32', via=self.fip)
        expected_calls = [mock.call(CONF.bgp_nic, ['192.168.1.10']),
                          mock.call(CONF.bgp_nic, ['192.168.1.11']),
                          mock.call(CONF.bgp_nic, ['192.168.1.13'])]
        mock_add_ips_dev.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(driver_utils, 'is_ipv6_gua')
    def test__process_lrp_port_gua(self, mock_ipv6_gua, mock_ip_version,
                                   mock_add_rule, mock_add_route,
                                   mock_add_ips_dev):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        CONF.set_override('expose_ipv6_gua_tenant_networks', True)
        self.addCleanup(CONF.clear_override, 'expose_ipv6_gua_tenant_networks')

        mock_ipv6_gua.return_value = True
        mock_ip_version.return_value = constants.IP_VERSION_6
        gateway = {}
        gateway['ips'] = ['{}/32'.format(self.fip),
                          '2003::1234:abcd:ffff:c0a8:102/128']
        gateway['provider_datapath'] = 'bc6780f4-9510-4270-b4d2-b8d5c6802713'
        gateway['subnets_datapath'] = {}
        gateway['subnets_cidr'] = []
        gateway['bridge_device'] = self.bridge
        gateway['bridge_vlan'] = 10
        self.bgp_driver.ovn_local_cr_lrps = {'gateway_port': gateway}

        router_port = fakes.create_object({
            'chassis': [],
            'mac': ['{} {}/128'.format(self.mac, self.ipv6)],
            'logical_port': 'lrp-fake-logical-port',
            'options': {'peer': 'fake-peer'}})

        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)

        dp_port0 = fakes.create_object({
            'name': 'fake-port-dp0',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': ['aa:bb:cc:dd:ee:ee 2002::1234:abcd:ffff:c0a8:111'],
            'chassis': 'fake-chassis1',
            'external_ids': {}})
        dp_port1 = fakes.create_object({
            'name': 'fake-port-dp1',
            'type': 'fake-type',
            'mac': ['aa:bb:cc:dd:ee:ee 2002::1234:abcd:ffff:c0a8:112'],
            'chassis': 'fake-chassis2',
            'external_ids': {}})
        dp_port2 = fakes.create_object({
            'name': 'fake-port-dp2',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': 'fake-chassis2',
            'external_ids': {}})
        dp_port3 = fakes.create_object({
            'name': 'fake-port-dp3',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': '',
            'up': [False],
            'external_ids': {}})
        dp_port4 = fakes.create_object({
            'name': 'fake-port-dp4',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': '',
            'up': [False],
            'external_ids': {
                constants.OVN_CIDRS_EXT_ID_KEY:
                    "2002::1234:abcd:ffff:c0a8:121/64"}})
        self.sb_idl.get_ports_on_datapath.return_value = [dp_port0, dp_port1,
                                                          dp_port2, dp_port3,
                                                          dp_port4]

        self.bgp_driver._process_lrp_port(router_port, 'gateway_port')

        # Assert that the add methods were called
        mock_ipv6_gua.assert_called_once_with('{}/128'.format(self.ipv6))
        mock_add_rule.assert_called_once_with(
            '{}/128'.format(self.ipv6), 'fake-table')
        mock_add_route.assert_called_once_with(
            mock.ANY, self.ipv6, 'fake-table', self.bridge,
            vlan=10, mask='128', via=self.fip)
        expected_calls = [mock.call(CONF.bgp_nic,
                                    ['2002::1234:abcd:ffff:c0a8:111']),
                          mock.call(CONF.bgp_nic,
                                    ['2002::1234:abcd:ffff:c0a8:121'])]
        mock_add_ips_dev.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(driver_utils, 'is_ipv6_gua')
    def test__process_lrp_port_not_gua(
            self, mock_ipv6_gua, mock_add_rule,
            mock_add_route, mock_add_ips_dev):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        CONF.set_override('expose_ipv6_gua_tenant_networks', True)
        self.addCleanup(CONF.clear_override, 'expose_ipv6_gua_tenant_networks')
        mock_ipv6_gua.return_value = False
        gateway = {}
        gateway['ips'] = ['{}/32'.format(self.fip),
                          '2003::1234:abcd:ffff:c0a8:102/128']
        gateway['provider_datapath'] = 'bc6780f4-9510-4270-b4d2-b8d5c6802713'
        gateway['subnets_datapath'] = {}
        gateway['subnets_cidr'] = []
        gateway['bridge_device'] = self.bridge
        gateway['bridge_vlan'] = 10
        self.bgp_driver.ovn_local_cr_lrps = {'gateway_port': gateway}

        router_port = fakes.create_object({
            'chassis': [],
            'mac': ['{} fdab:4ad8:e8fb:0:f816:3eff:fec6:469c/128'.format(
                self.mac)],
            'logical_port': 'lrp-fake-logical-port',
            'options': {'peer': 'fake-peer'}})

        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)

        self.bgp_driver._process_lrp_port(router_port, 'gateway_port')

        # Assert that the add methods were called
        mock_ipv6_gua.assert_called_once_with(
            'fdab:4ad8:e8fb:0:f816:3eff:fec6:469c/128')
        mock_add_rule.assert_not_called()
        mock_add_route.assert_not_called()
        mock_add_ips_dev.assert_not_called()

    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__process_lrp_port_invalid_ip(
            self, mock_ip_version, mock_add_rule, mock_add_route):
        mock_ip_version.return_value = constants.IP_VERSION_4
        gateway = {}
        gateway['ips'] = ['{}/32'.format(self.fip),
                          '2003::1234:abcd:ffff:c0a8:102/128']
        gateway['provider_datapath'] = 'bc6780f4-9510-4270-b4d2-b8d5c6802713'
        gateway['subnets_datapath'] = {}
        gateway['subnets_cidr'] = []
        gateway['bridge_device'] = self.bridge
        gateway['bridge_vlan'] = 10
        self.bgp_driver.ovn_local_cr_lrps = {'gateway_port': gateway}

        router_port = fakes.create_object({
            'chassis': [],
            'mac': ['{} {}/32'.format(self.mac, self.ipv4)],
            'logical_port': 'lrp-fake-logical-port',
            'options': {'peer': 'fake-peer'}})

        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)

        # Raise an exception on add_ip_rule()
        mock_add_rule.side_effect = agent_exc.InvalidPortIP(ip=self.ipv4)

        self.bgp_driver._process_lrp_port(router_port, 'gateway_port')

        mock_add_rule.assert_called_once_with(
            '{}/32'.format(self.ipv4), 'fake-table')
        # Assert that add_ip_route() was not called
        mock_add_route.assert_not_called()

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    def test__process_lrp_port_address_scopes(
            self, mock_add_rule, mock_add_route, mock_add_ips_dev):
        gateway = {}
        gateway['ips'] = ['{}/32'.format(self.fip),
                          '2003::1234:abcd:ffff:c0a8:102/128']
        gateway['provider_datapath'] = 'bc6780f4-9510-4270-b4d2-b8d5c6802713'
        gateway['subnets_datapath'] = {}
        gateway['subnets_cidr'] = []
        gateway['bridge_device'] = self.bridge
        gateway['bridge_vlan'] = 10
        self.bgp_driver.ovn_local_cr_lrps = {'gateway_port': gateway}
        mock_address_scope_allowed = mock.patch.object(
            self.bgp_driver, '_address_scope_allowed').start()
        mock_address_scope_allowed.return_value = False

        router_port = fakes.create_object({
            'chassis': [],
            'mac': ['{} {}/32'.format(self.mac, self.ipv4)],
            'logical_port': 'lrp-fake-logical-port',
            'options': {'peer': 'fake-peer'}})

        self.bgp_driver._process_lrp_port(router_port, 'gateway_port')

        # Assert that the add methods were called
        mock_add_rule.assert_not_called()
        mock_add_route.assert_not_called()
        mock_add_ips_dev.assert_not_called()

    def test__get_bridge_for_datapath(self):
        self.sb_idl.get_network_name_and_tag.return_value = (
            'fake-network', [10])
        ret = self.bgp_driver._get_bridge_for_datapath('fake-dp')
        self.assertEqual((self.bridge, 10), ret)

    def test__get_bridge_for_datapath_no_tag(self):
        self.sb_idl.get_network_name_and_tag.return_value = (
            'fake-network', None)
        ret = self.bgp_driver._get_bridge_for_datapath('fake-dp')
        self.assertEqual((self.bridge, None), ret)

    def test__get_bridge_for_datapath_no_network_name(self):
        self.sb_idl.get_network_name_and_tag.return_value = (None, None)
        ret = self.bgp_driver._get_bridge_for_datapath('fake-dp')
        self.assertEqual((None, None), ret)

    def test_expose_ovn_lb(self):
        mock_process_ovn_lb = mock.patch.object(
            self.bgp_driver, '_process_ovn_lb').start()
        self.bgp_driver.expose_ovn_lb('fake-ip', 'fake-row')
        mock_process_ovn_lb.assert_called_once_with(
            'fake-ip', 'fake-row', constants.EXPOSE)

    def test_withdraw_ovn_lb(self):
        mock_process_ovn_lb = mock.patch.object(
            self.bgp_driver, '_process_ovn_lb').start()
        self.bgp_driver.withdraw_ovn_lb('fake-ip', 'fake-row')
        mock_process_ovn_lb.assert_called_once_with(
            'fake-ip', 'fake-row', constants.WITHDRAW)

    def _test_process_ovn_lb(self, action, provider=False):
        mock_expose_remote_ip = mock.patch.object(
            self.bgp_driver, '_expose_remote_ip').start()
        mock_withdraw_remote_ip = mock.patch.object(
            self.bgp_driver, '_withdraw_remote_ip').start()

        self.sb_idl.is_provider_network.return_value = provider
        ip = 'fake-vip-ip'
        row = fakes.create_object({
            'logical_port': 'fake-vip',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'datapath': 'fake-provider-dp'})

        self.bgp_driver._process_ovn_lb(ip, row, action)

        if provider:
            mock_expose_remote_ip.assert_not_called()
            mock_withdraw_remote_ip.assert_not_called()
        else:
            if action == constants.EXPOSE:
                mock_expose_remote_ip.assert_called_once_with([ip], row)
                mock_withdraw_remote_ip.assert_not_called()

            elif action == constants.WITHDRAW:
                mock_expose_remote_ip.assert_not_called()
                mock_withdraw_remote_ip.assert_called_once_with([ip], row)
            else:
                mock_expose_remote_ip.assert_not_called()
                mock_withdraw_remote_ip.assert_not_called()

    def test__process_ovn_lb_expose_provider(self):
        self._test_process_ovn_lb(action=constants.EXPOSE, provider=True)

    def test__process_ovn_lb_expose_no_provider(self):
        self._test_process_ovn_lb(action=constants.EXPOSE)

    def test__process_ovn_lb_withdraw_provider(self):
        self._test_process_ovn_lb(action=constants.WITHDRAW, provider=True)

    def test__process_ovn_lb_withdraw_no_provider(self):
        self._test_process_ovn_lb(action=constants.WITHDRAW)

    def test__process_ovn_lb_unknown_action(self):
        self._test_process_ovn_lb(action="fake-action")

    def test__process_ovn_lb_datapath_exception(self):
        ip = 'fake-vip-ip'
        row = fakes.create_object({
            'logical_port': 'fake-vip',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'datapath': 'fake-provider-dp'})
        mock_expose_remote_ip = mock.patch.object(
            self.bgp_driver, '_expose_remote_ip').start()
        mock_withdraw_remote_ip = mock.patch.object(
            self.bgp_driver, '_withdraw_remote_ip').start()
        self.sb_idl.is_provider_network.side_effect = (
            agent_exc.DatapathNotFound(datapath=row.datapath))

        self.bgp_driver._process_ovn_lb(ip, row, mock.ANY)

        mock_expose_remote_ip.assert_not_called()
        mock_withdraw_remote_ip.assert_not_called()

    def test_expose_ovn_lb_on_provider(self):
        mock_expose_provider_port = mock.patch.object(
            self.bgp_driver, '_expose_provider_port').start()
        self.bgp_driver.expose_ovn_lb_on_provider(
            self.ipv4, 'ovn-lb-2', self.cr_lrp0)

        # Assert that the add methods were called
        mock_expose_provider_port.assert_called_once_with(
            [self.ipv4], self.bgp_driver.ovn_local_cr_lrps[self.cr_lrp0][
                'provider_datapath'], bridge_device=self.bridge,
            bridge_vlan=None)

    def test__expose_ovn_lb_on_provider_failure(self):
        mock_expose_provider_port = mock.patch.object(
            self.bgp_driver, '_expose_provider_port').start()
        mock_expose_provider_port.return_value = False
        ret = self.bgp_driver._expose_ovn_lb_on_provider(
            self.ipv4, self.loadbalancer_vip_port, self.cr_lrp0)

        # Assert that the add methods were called
        mock_expose_provider_port.assert_called_once_with(
            [self.ipv4], self.bgp_driver.ovn_local_cr_lrps[self.cr_lrp0][
                'provider_datapath'], bridge_device=self.bridge,
            bridge_vlan=None)
        self.assertEqual(False, ret)

    def test__expose_ovn_lb_on_provider_keyerror(self):
        mock_expose_provider_port = mock.patch.object(
            self.bgp_driver, '_expose_provider_port').start()
        ret = self.bgp_driver._expose_ovn_lb_on_provider(
            self.ipv4, self.loadbalancer_vip_port, 'wrong-cr-logical-port')

        # Assert that the add methods were called
        mock_expose_provider_port.assert_not_called()
        self.assertEqual(False, ret)

    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_ovn_lb_on_provider(
            self, mock_del_ip_dev, mock_del_rule, mock_del_route):
        self.bgp_driver.withdraw_ovn_lb_on_provider(
            self.loadbalancer_vip_port, self.cr_lrp0)

        # Assert that the del methods were called
        expected_calls = [mock.call(CONF.bgp_nic, [self.ipv4]),
                          mock.call(CONF.bgp_nic, [self.ipv6])]
        mock_del_ip_dev.assert_has_calls(expected_calls)

        expected_calls = [mock.call(self.ipv4, 'fake-table',
                                    dev='fake-bridge'),
                          mock.call(self.ipv6, 'fake-table',
                                    dev='fake-bridge')]
        mock_del_rule.assert_has_calls(expected_calls)

        expected_calls = [mock.call(mock.ANY, self.ipv4, 'fake-table',
                                    self.bridge, vlan=None),
                          mock.call(mock.ANY, self.ipv6, 'fake-table',
                                    self.bridge, vlan=None)]
        mock_del_route.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test__withdraw_ovn_lb_on_provider_keyerror(
            self, mock_del_ip_dev, mock_del_rule, mock_del_route):
        ret = self.bgp_driver._withdraw_ovn_lb_on_provider(
            self.loadbalancer_vip_port, 'wrong-cr-logical-port')

        # Assert that the del methods were called
        self.assertEqual(False, ret)
        mock_del_ip_dev.assert_not_called()
        mock_del_rule.assert_not_called()
        mock_del_route.assert_not_called()

    def test__withdraw_ovn_lb_on_provider_failure(self):
        mock_withdraw_provider_port = mock.patch.object(
            self.bgp_driver, '_withdraw_provider_port').start()
        mock_withdraw_provider_port.return_value = False
        ret = self.bgp_driver._withdraw_ovn_lb_on_provider(
            self.loadbalancer_vip_port, self.cr_lrp0)
        self.assertEqual(False, ret)

    @mock.patch.object(wire_utils, '_ensure_updated_mac_tweak_flows')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test_expose_ip_vm_on_provider_network(
            self, mock_add_ip_dev, mock_add_rule, mock_add_route,
            mock_ensure_mac_tweak):
        self.sb_idl.is_provider_network.return_value = True
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        row = fakes.create_object({
            'logical_port': 'fake-row',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.expose_ip(ips, row)

        # Assert that the add methods were called
        mock_ensure_mac_tweak.assert_called_once_with(mock.ANY, self.bridge,
                                                      {})
        mock_add_ip_dev.assert_called_once_with(
            CONF.bgp_nic, ips)

        expected_calls = [mock.call(self.ipv4, 'fake-table',
                                    dev='fake-bridge'),
                          mock.call(self.ipv6, 'fake-table',
                                    dev='fake-bridge')]
        mock_add_rule.assert_has_calls(expected_calls)

        expected_calls = [mock.call(mock.ANY, self.ipv4, 'fake-table',
                                    self.bridge, vlan=10),
                          mock.call(mock.ANY, self.ipv6, 'fake-table',
                                    self.bridge, vlan=10)]
        mock_add_route.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test__expose_ip_vm_on_provider_network_datapath_not_found(
            self, mock_add_ip_dev, mock_add_rule, mock_add_route):
        self.sb_idl.is_provider_network.side_effect = (
            agent_exc.DatapathNotFound(datapath="fake-dp"))
        row = fakes.create_object({
            'logical_port': 'fake-row',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        ret = self.bgp_driver._expose_ip(ips, row)

        # Assert that the add methods were called
        self.assertEqual([], ret)
        mock_add_ip_dev.assert_not_called()
        mock_add_rule.assert_not_called()
        mock_add_route.assert_not_called()

    @mock.patch.object(wire_utils, 'wire_provider_port')
    @mock.patch.object(bgp_utils, 'announce_ips')
    def test__expose_ip_vm_on_provider_network_expose_failure(
            self, mock_bgp_announce, mock_wire_port):
        self.sb_idl.is_provider_network.return_value = True
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        mock_wire_port.return_value = False
        row = fakes.create_object({
            'logical_port': 'fake-row',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        ret = self.bgp_driver._expose_ip(ips, row)

        # Assert that the add methods were called
        self.assertEqual([], ret)
        mock_wire_port.assert_called_once()
        mock_bgp_announce.assert_not_called()

    @mock.patch.object(wire_utils, '_ensure_updated_mac_tweak_flows')
    @mock.patch.object(linux_net, 'add_ndp_proxy')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test__expose_ip_virtual_port_on_provider_network(
            self, mock_add_ip_dev, mock_add_rule, mock_add_route,
            mock_ip_version, mock_add_ndp_proxy, mock_ensure_mac_tweak):
        self.sb_idl.is_provider_network.return_value = True
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        mock_ip_version.return_value = constants.IP_VERSION_6
        row = fakes.create_object({
            'logical_port': 'fake-row',
            'type': constants.OVN_VIRTUAL_VIF_PORT_TYPE,
            'datapath': 'fake-dp',
            'external_ids': {'neutron:cidrs': '{}/128'.format(self.ipv6)}})

        ips = [self.ipv4, self.ipv6]
        ret = self.bgp_driver._expose_ip(ips, row)

        # Assert that the add methods were called
        mock_ensure_mac_tweak.assert_called_once_with(mock.ANY, self.bridge,
                                                      {})
        self.assertEqual(ips, ret)
        mock_add_ip_dev.assert_called_once_with(
            CONF.bgp_nic, ips)

        expected_calls = [mock.call(self.ipv4, 'fake-table',
                                    dev='fake-bridge'),
                          mock.call(self.ipv6, 'fake-table',
                                    dev='fake-bridge')]
        mock_add_rule.assert_has_calls(expected_calls)

        expected_calls = [mock.call(mock.ANY, self.ipv4, 'fake-table',
                                    self.bridge, vlan=10),
                          mock.call(mock.ANY, self.ipv6, 'fake-table',
                                    self.bridge, vlan=10)]
        mock_add_route.assert_has_calls(expected_calls)
        mock_add_ndp_proxy.assert_called_once_with(
            '{}/128'.format(self.ipv6), self.bridge, 10)

    @mock.patch.object(linux_net, 'add_ndp_proxy')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test__expose_ip_virtual_port_on_provider_network_expose_failure(
            self, mock_add_ip_dev, mock_add_rule, mock_add_route,
            mock_ip_version, mock_add_ndp_proxy):
        self.sb_idl.is_provider_network.return_value = True
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (None, None)
        mock_ip_version.return_value = constants.IP_VERSION_6
        row = fakes.create_object({
            'logical_port': 'fake-row',
            'type': constants.OVN_VIRTUAL_VIF_PORT_TYPE,
            'datapath': 'fake-dp',
            'external_ids': {'neutron:cidrs': '{}/128'.format(self.ipv6)}})

        ips = [self.ipv4, self.ipv6]
        ret = self.bgp_driver._expose_ip(ips, row)

        # Assert that the add methods were called
        self.assertEqual([], ret)
        mock_add_ip_dev.assert_not_called()
        mock_add_rule.assert_not_called()
        mock_add_route.assert_not_called()
        mock_add_ndp_proxy.assert_not_called()

    @mock.patch.object(wire_utils, '_ensure_updated_mac_tweak_flows')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test__expose_ip_vm_with_fip(
            self, mock_add_ip_dev, mock_add_rule, mock_add_route,
            mock_ensure_mac_tweak):
        self.sb_idl.is_provider_network.side_effect = [False, True]
        self.sb_idl.get_fip_associated.return_value = (
            self.fip, 'fake-dp')
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        row = fakes.create_object({
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'logical_port': 'fake-logical-port',
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        ret = self.bgp_driver._expose_ip(ips, row)

        # Assert that the add methods were called
        self.assertEqual([self.fip], ret)
        mock_ensure_mac_tweak.assert_called_once_with(mock.ANY, self.bridge,
                                                      {})
        mock_add_ip_dev.assert_called_once_with(
            CONF.bgp_nic, [self.fip])
        mock_add_rule.assert_called_once_with(
            self.fip, 'fake-table', dev='fake-bridge')
        mock_add_route.assert_called_once_with(
            mock.ANY, self.fip, 'fake-table', self.bridge, vlan=10)

    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test__expose_ip_vm_with_fip_no_provider(
            self, mock_add_ip_dev, mock_add_rule, mock_add_route):
        self.sb_idl.is_provider_network.side_effect = [False, False]
        self.sb_idl.get_fip_associated.return_value = (
            self.fip, 'fake-dp')
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        row = fakes.create_object({
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'logical_port': 'fake-logical-port',
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        ret = self.bgp_driver._expose_ip(ips, row)

        # Assert that the add methods were called
        self.assertEqual([], ret)
        mock_add_ip_dev.assert_not_called()
        mock_add_rule.assert_not_called()
        mock_add_route.assert_not_called()

    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test__expose_ip_vm_with_fip_no_fip_address(
            self, mock_add_ip_dev, mock_add_rule, mock_add_route):
        self.sb_idl.is_provider_network.return_value = False
        self.sb_idl.get_fip_associated.return_value = (None, None)
        row = fakes.create_object({
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'logical_port': 'fake-logical-port',
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        ret = self.bgp_driver._expose_ip(ips, row)

        # Assert that the add methods were not called
        self.assertEqual([], ret)
        mock_add_ip_dev.assert_not_called()
        mock_add_rule.assert_not_called()
        mock_add_route.assert_not_called()

    @mock.patch.object(wire_utils, '_ensure_updated_mac_tweak_flows')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test__expose_ip_fip_association_to_vm(
            self, mock_add_ip_dev, mock_add_rule, mock_add_route,
            mock_ensure_mac_tweak):
        self.sb_idl.is_provider_network.return_value = True
        self.sb_idl.is_port_on_chassis.return_value = True
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        row = fakes.create_object({
            'type': constants.OVN_PATCH_VIF_PORT_TYPE,
            'logical_port': 'fake-logical-port',
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        ret = self.bgp_driver._expose_ip(ips, row,
                                         associated_port=self.cr_lrp0)

        # Assert that the add methods were called
        self.assertEqual(ips, ret)
        mock_ensure_mac_tweak.assert_called_once_with(mock.ANY, self.bridge,
                                                      {})
        mock_add_ip_dev.assert_called_once_with(
            CONF.bgp_nic, ips)

        expected_calls = [mock.call(self.ipv4, 'fake-table',
                                    dev='fake-bridge'),
                          mock.call(self.ipv6, 'fake-table',
                                    dev='fake-bridge')]
        mock_add_rule.assert_has_calls(expected_calls)

        expected_calls = [mock.call(mock.ANY, self.ipv4, 'fake-table',
                                    self.bridge, vlan=10),
                          mock.call(mock.ANY, self.ipv6, 'fake-table',
                                    self.bridge, vlan=10)]
        mock_add_route.assert_has_calls(expected_calls)

    @mock.patch.object(wire_utils, '_ensure_updated_mac_tweak_flows')
    @mock.patch.object(linux_net, 'add_ndp_proxy')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test__expose_ip_chassisredirect_port(
            self, mock_add_ip_dev, mock_add_rule, mock_add_route,
            mock_ip_version, mock_ndp_proxy, mock_ensure_mac_tweak):
        self.sb_idl.get_provider_datapath_from_cr_lrp.return_value = (
            'fake-provider-dp')
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)

        lrp0 = fakes.create_object({
            'logical_port': self.lrp0,
            'chassis': '',
            'mac': ["fa:16:3e:50:ec:81 192.168.1.1/24"],
            'options': {}})
        lrp1 = fakes.create_object({'logical_port': 'lrp-1',
                                    'chassis': 'fake-chassis',
                                    'options': {}})
        lrp2 = fakes.create_object({'logical_port': 'fake-lrp',
                                    'chassis': '',
                                    'options': {}})
        self.sb_idl.get_lrp_ports_for_router.return_value = [lrp0, lrp1, lrp2]

        self.sb_idl.get_port_datapath.return_value = 'fake-lrp-dp'

        self.sb_idl.get_cr_lrp_nat_addresses_info.return_value = (
            [], self.cr_lrp0)

        mock_process_lrp_port = mock.patch.object(
            self.bgp_driver, '_process_lrp_port').start()

        mock_ip_version.side_effect = (constants.IP_VERSION_4,
                                       constants.IP_VERSION_6)
        row = fakes.create_object({
            'type': constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            'logical_port': self.cr_lrp0,
            'mac': ['{} {} {}'.format(self.mac, self.ipv4, self.ipv6)],
            'datapath': 'fake-router-dp'})

        ovn_lb_vip = '172.24.4.5'
        ovn_lbs = {'fake-vip-port': ovn_lb_vip}
        self.sb_idl.get_provider_ovn_lbs_on_cr_lrp.return_value = (
            ovn_lbs)
        mock_expose_ovn_lb = mock.patch.object(
            self.bgp_driver, '_expose_ovn_lb_on_provider').start()

        ips = [self.ipv4, self.ipv6]
        ret = self.bgp_driver._expose_ip(ips, row)

        # Assert that the add methods were called
        self.assertEqual(ips, ret)
        mock_ensure_mac_tweak.assert_called_once_with(mock.ANY, self.bridge,
                                                      {})
        mock_add_ip_dev.assert_called_once_with(
            CONF.bgp_nic, ips)

        expected_calls = [mock.call(self.ipv4, 'fake-table',
                          dev='{}.{}'.format(self.bridge, 10),
                          lladdr=self.mac),
                          mock.call(self.ipv6, 'fake-table',
                          dev='{}.{}'.format(self.bridge, 10),
                          lladdr=self.mac)]
        mock_add_rule.assert_has_calls(expected_calls)

        expected_calls = [mock.call(mock.ANY, self.ipv4, 'fake-table',
                                    self.bridge, vlan=10),
                          mock.call(mock.ANY, self.ipv6, 'fake-table',
                                    self.bridge, vlan=10)]
        mock_add_route.assert_has_calls(expected_calls)

        mock_ndp_proxy.assert_called_once_with(self.ipv6, self.bridge, 10)

        expected_calls = [mock.call(lrp0, self.cr_lrp0),
                          mock.call(lrp1, self.cr_lrp0),
                          mock.call(lrp2, self.cr_lrp0)]
        mock_process_lrp_port.assert_has_calls(expected_calls)
        mock_expose_ovn_lb.assert_called_once_with(
            ovn_lb_vip, 'fake-vip-port', self.cr_lrp0)

    @mock.patch.object(linux_net, 'add_ndp_proxy')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test__expose_ip_chassisredirect_port_no_datapath(
            self, mock_add_ip_dev, mock_add_rule, mock_add_route,
            mock_ndp_proxy):
        self.sb_idl.get_provider_datapath_from_cr_lrp.return_value = None

        row = fakes.create_object({
            'type': constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            'logical_port': self.cr_lrp0,
            'mac': ['{} {} {}'.format(self.mac, self.ipv4, self.ipv6)],
            'datapath': 'fake-router-dp'})

        ips = [self.ipv4, self.ipv6]
        ret = self.bgp_driver._expose_ip(ips, row)

        # Assert that the add methods were called
        self.assertEqual([], ret)
        mock_add_ip_dev.assert_not_called()
        mock_add_rule.assert_not_called()
        mock_add_route.assert_not_called()
        mock_ndp_proxy.assert_not_called()

    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_ip_vm_on_provider_network(
            self, mock_del_ip_dev, mock_del_rule, mock_del_route):
        self.sb_idl.is_provider_network.return_value = True
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        row = fakes.create_object({
            'logical_port': 'fake-row',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.withdraw_ip(ips, row)

        # Assert that the del methods were called
        mock_del_ip_dev.assert_called_once_with(
            CONF.bgp_nic, ips)

        expected_calls = [mock.call(self.ipv4, 'fake-table',
                                    dev='fake-bridge'),
                          mock.call(self.ipv6, 'fake-table',
                                    dev='fake-bridge')]
        mock_del_rule.assert_has_calls(expected_calls)

        expected_calls = [mock.call(mock.ANY, self.ipv4, 'fake-table',
                                    self.bridge, vlan=10),
                          mock.call(mock.ANY, self.ipv6, 'fake-table',
                                    self.bridge, vlan=10)]
        mock_del_route.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'del_ndp_proxy')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_ip_virtual_port_on_provider_network(
            self, mock_del_ip_dev, mock_del_rule, mock_del_route,
            mock_ip_version, mock_del_ndp_proxy):
        self.sb_idl.is_provider_network.return_value = True
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        row = fakes.create_object({
            'logical_port': 'fake-row',
            'type': constants.OVN_VIRTUAL_VIF_PORT_TYPE,
            'datapath': 'fake-dp',
            'external_ids': {'neutron:cidrs': '{}/128'.format(self.ipv6)}})

        ips = [self.ipv4, self.ipv6]
        self.sb_idl.get_virtual_ports_on_datapath_by_chassis.return_value = []
        mock_ip_version.return_value = constants.IP_VERSION_6

        self.bgp_driver.withdraw_ip(ips, row)

        # Assert that the del methods were called
        mock_del_ip_dev.assert_called_once_with(
            CONF.bgp_nic, ips)

        expected_calls = [mock.call(self.ipv4, 'fake-table',
                                    dev='fake-bridge'),
                          mock.call(self.ipv6, 'fake-table',
                                    dev='fake-bridge')]
        mock_del_rule.assert_has_calls(expected_calls)

        expected_calls = [mock.call(mock.ANY, self.ipv4, 'fake-table',
                                    self.bridge, vlan=10),
                          mock.call(mock.ANY, self.ipv6, 'fake-table',
                                    self.bridge, vlan=10)]
        mock_del_route.assert_has_calls(expected_calls)
        mock_del_ndp_proxy.assert_called_once_with(
            '{}/128'.format(self.ipv6), self.bridge, 10)

    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_ip_vm_with_fip(
            self, mock_del_ip_dev, mock_del_rule, mock_del_route):
        self.sb_idl.is_provider_network.side_effect = [False, True]
        self.sb_idl.get_fip_associated.return_value = (
            self.fip, 'fake-dp')
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        row = fakes.create_object({
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'logical_port': 'fake-logical-port',
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.withdraw_ip(ips, row)

        # Assert that the del methods were called
        mock_del_ip_dev.assert_called_once_with(
            CONF.bgp_nic, [self.fip])
        mock_del_rule.assert_called_once_with(
            self.fip, 'fake-table', dev='fake-bridge')
        mock_del_route.assert_called_once_with(
            mock.ANY, self.fip, 'fake-table', self.bridge, vlan=10)

    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_ip_vm_with_fip_no_fip_address(
            self, mock_del_ip_dev, mock_del_rule, mock_del_route):
        self.sb_idl.is_provider_network.return_value = False
        self.sb_idl.get_fip_associated.return_value = (None, None)
        row = fakes.create_object({
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'logical_port': 'fake-logical-port',
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.withdraw_ip(ips, row)

        # Assert that the del methods were not called
        mock_del_ip_dev.assert_not_called()
        mock_del_rule.assert_not_called()
        mock_del_route.assert_not_called()

    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_ip_fip_association_to_vm(
            self, mock_del_ip_dev, mock_del_rule, mock_del_route):
        self.sb_idl.is_provider_network.return_value = True
        self.sb_idl.is_port_on_chassis.return_value = True
        mock_get_bridge = mock.patch.object(
            self.bgp_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, 10)
        row = fakes.create_object({
            'type': constants.OVN_PATCH_VIF_PORT_TYPE,
            'logical_port': 'fake-logical-port',
            'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.withdraw_ip(ips, row, associated_port=self.cr_lrp0)

        # Assert that the del methods were called
        mock_del_ip_dev.assert_called_once_with(
            CONF.bgp_nic, ips)

        expected_calls = [mock.call(self.ipv4, 'fake-table',
                                    dev='fake-bridge'),
                          mock.call(self.ipv6, 'fake-table',
                                    dev='fake-bridge')]
        mock_del_rule.assert_has_calls(expected_calls)

        expected_calls = [mock.call(mock.ANY, self.ipv4, 'fake-table',
                                    self.bridge, vlan=10),
                          mock.call(mock.ANY, self.ipv6, 'fake-table',
                                    self.bridge, vlan=10)]
        mock_del_route.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'del_ndp_proxy')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_ip_chassisredirect_port(
            self, mock_del_ip_dev, mock_del_rule, mock_del_route,
            mock_ip_version, mock_ndp_proxy):
        mock_withdraw_lrp_port = mock.patch.object(
            self.bgp_driver, '_withdraw_lrp_port').start()

        mock_ip_version.side_effect = (constants.IP_VERSION_4,
                                       constants.IP_VERSION_6,
                                       constants.IP_VERSION_4,
                                       constants.IP_VERSION_6,
                                       constants.IP_VERSION_6)
        row = fakes.create_object({
            'type': constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            'logical_port': self.cr_lrp0,
            'mac': ['{} {} {}'.format(self.mac, self.ipv4, self.ipv6)],
            'datapath': 'fake-router-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.withdraw_ip(ips, row)

        # Assert that the del methods were called
        mock_del_ip_dev.assert_called_once_with(
            CONF.bgp_nic, ips)

        expected_calls = [mock.call('{}/32'.format(self.ipv4), 'fake-table',
                          dev=self.bridge, lladdr=self.mac),
                          mock.call('{}/128'.format(self.ipv6), 'fake-table',
                          dev=self.bridge, lladdr=self.mac)]
        mock_del_rule.assert_has_calls(expected_calls)

        expected_calls = [mock.call(mock.ANY, self.ipv4, 'fake-table',
                                    self.bridge, vlan=None),
                          mock.call(mock.ANY, self.ipv6, 'fake-table',
                                    self.bridge, vlan=None)]
        mock_del_route.assert_has_calls(expected_calls)

        mock_ndp_proxy.assert_called_once_with(self.ipv6, self.bridge, None)

        mock_withdraw_lrp_port.assert_called_once_with(
            '192.168.1.1/24', None, self.cr_lrp0)

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test_expose_remote_ip(self, mock_add_ip_dev):
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.bgp_driver.ovn_local_lrps = {lrp: 'fake-cr-lrp'}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.expose_remote_ip(ips, row)

        mock_add_ip_dev.assert_called_once_with(CONF.bgp_nic, ips)

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test_expose_remote_ip_is_provider_network(self, mock_add_ip_dev):
        self.sb_idl.is_provider_network.return_value = True
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.expose_remote_ip(ips, row)

        mock_add_ip_dev.assert_not_called()

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test_expose_remote_ip_not_local(self, mock_add_ip_dev):
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.bgp_driver.ovn_local_lrps = {}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.expose_remote_ip(ips, row)

        mock_add_ip_dev.assert_not_called()

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(driver_utils, 'is_ipv6_gua')
    def test_expose_remote_ip_gua(self, mock_ipv6_gua, mock_add_ip_dev):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        CONF.set_override('expose_ipv6_gua_tenant_networks', True)
        self.addCleanup(CONF.clear_override, 'expose_ipv6_gua_tenant_networks')
        mock_ipv6_gua.side_effect = [False, True]
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.bgp_driver.ovn_local_lrps = {lrp: 'fake-cr-lrp'}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.expose_remote_ip(ips, row)

        mock_add_ip_dev.assert_called_once_with(CONF.bgp_nic, [self.ipv6])

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(driver_utils, 'is_ipv6_gua')
    def test_expose_remote_ip_not_gua(self, mock_ipv6_gua, mock_add_ip_dev):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        CONF.set_override('expose_ipv6_gua_tenant_networks', True)
        self.addCleanup(CONF.clear_override, 'expose_ipv6_gua_tenant_networks')
        mock_ipv6_gua.side_effect = [False, False]
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.bgp_driver.ovn_local_lrps = {lrp: 'fake-cr-lrp'}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, 'fdab:4ad8:e8fb:0:f816:3eff:fec6:469c']
        self.bgp_driver.expose_remote_ip(ips, row)

        mock_add_ip_dev.assert_not_called()

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test_expose_remote_ip_address_scope(self, mock_add_ip_dev):
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.bgp_driver.ovn_local_lrps = {lrp: 'fake-cr-lrp'}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        mock_address_scope_allowed = mock.patch.object(
            self.bgp_driver, '_address_scope_allowed').start()
        mock_address_scope_allowed.side_effect = [False, True]

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.expose_remote_ip(ips, row)

        mock_add_ip_dev.assert_called_once_with(CONF.bgp_nic, [self.ipv6])

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_remote_ip(self, mock_del_ip_dev):
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.bgp_driver.ovn_local_lrps = {lrp: 'fake-cr-lrp'}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.withdraw_remote_ip(ips, row)

        mock_del_ip_dev.assert_called_once_with(CONF.bgp_nic, ips)

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_remote_ip_is_provider_network(self, mock_del_ip_dev):
        self.sb_idl.is_provider_network.return_value = True
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.withdraw_remote_ip(ips, row)

        mock_del_ip_dev.assert_not_called()

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_remote_ip_not_local(self, mock_del_ip_dev):
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.bgp_driver.ovn_local_lrps = {}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.withdraw_remote_ip(ips, row)

        mock_del_ip_dev.assert_not_called()

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    @mock.patch.object(driver_utils, 'is_ipv6_gua')
    def test_withdraw_remote_ip_gua(self, mock_ipv6_gua, mock_del_ip_dev):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        CONF.set_override('expose_ipv6_gua_tenant_networks', True)
        self.addCleanup(CONF.clear_override, 'expose_ipv6_gua_tenant_networks')
        mock_ipv6_gua.side_effect = [False, True]
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.bgp_driver.ovn_local_lrps = {lrp: 'fake-cr-lrp'}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.withdraw_remote_ip(ips, row)

        mock_del_ip_dev.assert_called_once_with(CONF.bgp_nic, [self.ipv6])

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    @mock.patch.object(driver_utils, 'is_ipv6_gua')
    def test_withdraw_remote_ip_not_gua(self, mock_ipv6_gua, mock_del_ip_dev):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        CONF.set_override('expose_ipv6_gua_tenant_networks', True)
        self.addCleanup(CONF.clear_override, 'expose_ipv6_gua_tenant_networks')
        mock_ipv6_gua.side_effect = [False, False]
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.bgp_driver.ovn_local_lrps = {lrp: 'fake-cr-lrp'}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, 'fdab:4ad8:e8fb:0:f816:3eff:fec6:469c']
        self.bgp_driver.withdraw_remote_ip(ips, row)

        mock_del_ip_dev.assert_not_called()

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_remote_ip_address_scope(self, mock_del_ip_dev):
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.bgp_driver.ovn_local_lrps = {lrp: 'fake-cr-lrp'}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        mock_address_scope_allowed = mock.patch.object(
            self.bgp_driver, '_address_scope_allowed').start()
        mock_address_scope_allowed.side_effect = [False, True]

        ips = [self.ipv4, self.ipv6]
        self.bgp_driver.withdraw_remote_ip(ips, row)

        mock_del_ip_dev.assert_called_once_with(CONF.bgp_nic, [self.ipv6])

    @mock.patch.object(linux_net, 'get_ip_version')
    def test__expose_cr_lrp_port(self, mock_ip_version):
        mock_expose_provider_port = mock.patch.object(
            self.bgp_driver, '_expose_provider_port').start()
        mock_process_lrp_port = mock.patch.object(
            self.bgp_driver, '_process_lrp_port').start()
        mock_expose_ovn_lb = mock.patch.object(
            self.bgp_driver, '_expose_ovn_lb_on_provider').start()
        ips = [self.ipv4, self.ipv6]
        mock_ip_version.side_effect = [constants.IP_VERSION_4,
                                       constants.IP_VERSION_6]

        dp_port0 = mock.Mock()
        self.sb_idl.get_lrp_ports_for_router.return_value = [dp_port0]
        ovn_lbs = {'fake-vip-port': 'fake-vip-ip'}
        self.sb_idl.get_provider_ovn_lbs_on_cr_lrp.return_value = (
            ovn_lbs)

        ips_without_mask = [ip.split("/")[0] for ip in ips]
        self.sb_idl.get_cr_lrp_nat_addresses_info.return_value = (
            [ips_without_mask[0]], self.cr_lrp0)

        self.bgp_driver._expose_cr_lrp_port(
            ips, self.mac, self.bridge, None, router_datapath='fake-router-dp',
            provider_datapath='fake-provider-dp', cr_lrp_port=self.cr_lrp0)

        mock_expose_provider_port.assert_called_once_with(
            ips_without_mask, 'fake-provider-dp', self.bridge, None,
            lladdr=self.mac, proxy_cidrs=ips)
        mock_process_lrp_port.assert_called_once_with(dp_port0, self.cr_lrp0)
        mock_expose_ovn_lb.assert_called_once_with(
            'fake-vip-ip', 'fake-vip-port', self.cr_lrp0)

    def test__expose_cr_lrp_port_failure(self):
        mock_expose_provider_port = mock.patch.object(
            self.bgp_driver, '_expose_provider_port').start()
        mock_expose_provider_port.return_value = False
        mock_process_lrp_port = mock.patch.object(
            self.bgp_driver, '_process_lrp_port').start()
        mock_expose_ovn_lb = mock.patch.object(
            self.bgp_driver, '_expose_ovn_lb_on_provider').start()
        ips = [self.ipv4, self.ipv6]

        ret = self.bgp_driver._expose_cr_lrp_port(
            ips, self.mac, self.bridge, None, router_datapath='fake-router-dp',
            provider_datapath='fake-provider-dp', cr_lrp_port=self.cr_lrp0)

        self.assertEqual(False, ret)

        ips_without_mask = [ip.split("/")[0] for ip in ips]
        mock_expose_provider_port.assert_called_once_with(
            ips_without_mask, 'fake-provider-dp', self.bridge, None,
            lladdr=self.mac, proxy_cidrs=ips)
        mock_process_lrp_port.assert_not_called()
        mock_expose_ovn_lb.assert_not_called()

    @mock.patch.object(linux_net, 'get_ip_version')
    def test__withdraw_cr_lrp_port(self, mock_ip_version):
        mock_withdraw_provider_port = mock.patch.object(
            self.bgp_driver, '_withdraw_provider_port').start()
        mock_withdraw_lrp_port = mock.patch.object(
            self.bgp_driver, '_withdraw_lrp_port').start()
        mock_withdraw_ovn_lb_on_provider = mock.patch.object(
            self.bgp_driver, '_withdraw_ovn_lb_on_provider').start()

        ips = [self.ipv4, self.ipv6]
        mock_ip_version.side_effect = [constants.IP_VERSION_4,
                                       constants.IP_VERSION_6]
        ovn_lb_vip_port = mock.Mock()
        gateway = {
            'ips': ips,
            'provider_datapath': 'fake-provider-dp',
            'subnets_cidr': ['192.168.1.1/24'],
            'bridge_device': self.bridge,
            'bridge_vlan': 10,
            'mac': self.mac,
            'provider_ovn_lbs': [ovn_lb_vip_port]}
        self.bgp_driver.ovn_local_cr_lrps = {'gateway_port': gateway}

        self.bgp_driver._withdraw_cr_lrp_port(
            ips, self.mac, self.bridge, 10,
            provider_datapath='fake-provider-dp', cr_lrp_port='gateway_port')

        ips_without_mask = [ip.split("/")[0] for ip in ips]
        mock_withdraw_provider_port.assert_called_once_with(
            ips_without_mask, 'fake-provider-dp', bridge_device=self.bridge,
            bridge_vlan=10, lladdr=self.mac, proxy_cidrs=[self.ipv6])
        mock_withdraw_lrp_port.assert_called_once_with('192.168.1.1/24', None,
                                                       'gateway_port')
        mock_withdraw_ovn_lb_on_provider.assert_called_once_with(
            ovn_lb_vip_port, 'gateway_port')

    @mock.patch.object(linux_net, 'get_ip_version')
    def test__withdraw_cr_lrp_port_withdraw_failure(self, mock_ip_version):
        mock_withdraw_provider_port = mock.patch.object(
            self.bgp_driver, '_withdraw_provider_port').start()
        mock_withdraw_provider_port.return_value = False
        mock_withdraw_lrp_port = mock.patch.object(
            self.bgp_driver, '_withdraw_lrp_port').start()
        mock_withdraw_ovn_lb_on_provider = mock.patch.object(
            self.bgp_driver, '_withdraw_ovn_lb_on_provider').start()

        ips = [self.ipv4, self.ipv6]
        mock_ip_version.side_effect = [constants.IP_VERSION_4,
                                       constants.IP_VERSION_6]
        ovn_lb_vip_port = mock.Mock()
        gateway = {
            'ips': ips,
            'provider_datapath': 'fake-provider-dp',
            'subnets_cidr': ['192.168.1.1/24'],
            'bridge_device': self.bridge,
            'bridge_vlan': 10,
            'mac': self.mac,
            'provider_ovn_lbs': [ovn_lb_vip_port]}
        self.bgp_driver.ovn_local_cr_lrps = {'gateway_port': gateway}

        ret = self.bgp_driver._withdraw_cr_lrp_port(
            ips, self.mac, self.bridge, 10,
            provider_datapath='fake-provider-dp', cr_lrp_port='gateway_port')

        self.assertEqual(False, ret)
        ips_without_mask = [ip.split("/")[0] for ip in ips]
        mock_withdraw_provider_port.assert_called_once_with(
            ips_without_mask, 'fake-provider-dp', bridge_device=self.bridge,
            bridge_vlan=10, lladdr=self.mac, proxy_cidrs=[self.ipv6])
        mock_withdraw_lrp_port.assert_not_called()
        mock_withdraw_ovn_lb_on_provider.assert_not_called()

    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__expose_lrp_port(
            self, mock_ip_version, mock_add_rule, mock_add_route):
        mock_ip_version.return_value = constants.IP_VERSION_4
        dp_port0 = fakes.create_object({
            'name': 'fake-port-dp0',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.10 192.168.1.11'],
            'chassis': 'fake-chassis1'})
        dp_port1 = fakes.create_object({
            'name': 'fake-port-dp1',
            'type': 'fake-type',
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.12 192.168.1.13'],
            'chassis': 'fake-chassis2'})
        dp_port2 = fakes.create_object({
            'name': 'fake-port-dp1',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': 'fake-chassis2'})
        self.sb_idl.get_ports_on_datapath.return_value = [dp_port0, dp_port1,
                                                          dp_port2]
        mock_expose_tenant_port = mock.patch.object(
            self.bgp_driver, '_expose_tenant_port').start()

        self.bgp_driver._expose_lrp_port(
            '{}/32'.format(self.ipv4), self.lrp0, self.cr_lrp0, 'fake-lrp-dp')

        mock_add_rule.assert_called_once_with(
            '{}/32'.format(self.ipv4), 'fake-table')
        mock_add_route.assert_called_once_with(
            mock.ANY, self.ipv4, 'fake-table', self.bridge, vlan=None,
            mask='32', via=self.fip)
        expected_calls = [
            mock.call(dp_port0, ip_version=constants.IP_VERSION_4,
                      exposed_ips=None, ovn_ip_rules=None),
            mock.call(dp_port1, ip_version=constants.IP_VERSION_4,
                      exposed_ips=None, ovn_ip_rules=None),
            mock.call(dp_port2, ip_version=constants.IP_VERSION_4,
                      exposed_ips=None, ovn_ip_rules=None)]
        mock_expose_tenant_port.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__expose_lrp_port_invalid_ip(
            self, mock_ip_version, mock_add_rule, mock_add_route):
        mock_ip_version.return_value = constants.IP_VERSION_4
        dp_port0 = fakes.create_object({
            'name': 'fake-port-dp0',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.10 192.168.1.11'],
            'chassis': 'fake-chassis1'})
        dp_port1 = fakes.create_object({
            'name': 'fake-port-dp1',
            'type': 'fake-type',
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.12 192.168.1.13'],
            'chassis': 'fake-chassis2'})
        dp_port2 = fakes.create_object({
            'name': 'fake-port-dp1',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': 'fake-chassis2'})
        self.sb_idl.get_ports_on_datapath.return_value = [dp_port0, dp_port1,
                                                          dp_port2]
        mock_expose_tenant_port = mock.patch.object(
            self.bgp_driver, '_expose_tenant_port').start()

        mock_add_rule.side_effect = agent_exc.InvalidPortIP(ip=self.ipv4)

        self.bgp_driver._expose_lrp_port(
            '{}/32'.format(self.ipv4), self.lrp0, self.cr_lrp0, 'fake-lrp-dp')

        mock_add_rule.assert_called_once_with(
            '{}/32'.format(self.ipv4), 'fake-table')
        mock_add_route.assert_not_called()
        mock_expose_tenant_port.assert_not_called()

    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(driver_utils, 'is_ipv6_gua')
    def test__expose_lrp_port_gua(
            self, mock_ipv6_gua, mock_ip_version, mock_add_rule,
            mock_add_route):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        CONF.set_override('expose_ipv6_gua_tenant_networks', True)
        self.addCleanup(CONF.clear_override, 'expose_ipv6_gua_tenant_networks')
        mock_ipv6_gua.return_value = True
        mock_ip_version.return_value = constants.IP_VERSION_6
        dp_port0 = fakes.create_object({
            'name': 'fake-port-dp0',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': ['aa:bb:cc:dd:ee:ee 2002::1234:abcd:ffff:c0a8:111'],
            'chassis': 'fake-chassis1'})
        dp_port1 = fakes.create_object({
            'name': 'fake-port-dp1',
            'type': 'fake-type',
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.12 192.168.1.13'],
            'chassis': 'fake-chassis2'})
        dp_port2 = fakes.create_object({
            'name': 'fake-port-dp1',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': 'fake-chassis2'})
        self.sb_idl.get_ports_on_datapath.return_value = [dp_port0, dp_port1,
                                                          dp_port2]
        mock_expose_tenant_port = mock.patch.object(
            self.bgp_driver, '_expose_tenant_port').start()

        self.bgp_driver._expose_lrp_port(
            '{}/128'.format(self.ipv6), self.lrp0, self.cr_lrp0, 'fake-lrp-dp')

        mock_add_rule.assert_called_once_with(
            '{}/128'.format(self.ipv6), 'fake-table')
        mock_add_route.assert_called_once_with(
            mock.ANY, self.ipv6, 'fake-table', self.bridge, vlan=None,
            mask='128', via=self.fip)
        expected_calls = [
            mock.call(dp_port0, ip_version=constants.IP_VERSION_6,
                      exposed_ips=None, ovn_ip_rules=None),
            mock.call(dp_port1, ip_version=constants.IP_VERSION_6,
                      exposed_ips=None, ovn_ip_rules=None),
            mock.call(dp_port2, ip_version=constants.IP_VERSION_6,
                      exposed_ips=None, ovn_ip_rules=None)]
        mock_expose_tenant_port.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'add_ip_rule')
    @mock.patch.object(driver_utils, 'is_ipv6_gua')
    def test__expose_lrp_port_no_gua(
            self, mock_ipv6_gua, mock_add_rule, mock_add_route):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        CONF.set_override('expose_ipv6_gua_tenant_networks', True)
        self.addCleanup(CONF.clear_override, 'expose_ipv6_gua_tenant_networks')
        mock_ipv6_gua.return_value = False
        dp_port0 = fakes.create_object({
            'name': 'fake-port-dp0',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': ['aa:bb:cc:dd:ee:ee 2002::1234:abcd:ffff:c0a8:111'],
            'chassis': 'fake-chassis1'})
        dp_port1 = fakes.create_object({
            'name': 'fake-port-dp1',
            'type': 'fake-type',
            'mac': ['aa:bb:cc:dd:ee:ee 192.168.1.12 192.168.1.13'],
            'chassis': 'fake-chassis2'})
        dp_port2 = fakes.create_object({
            'name': 'fake-port-dp1',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'mac': [],
            'chassis': 'fake-chassis2'})
        self.sb_idl.get_ports_on_datapath.return_value = [dp_port0, dp_port1,
                                                          dp_port2]
        mock_expose_tenant_port = mock.patch.object(
            self.bgp_driver, '_expose_tenant_port').start()

        self.bgp_driver._expose_lrp_port(
            'fdab:4ad8:e8fb:0:f816:3eff:fec6:469c/128', self.lrp0,
            self.cr_lrp0, 'fake-lrp-dp')

        mock_add_rule.assert_not_called()
        mock_add_route.assert_not_called()
        mock_expose_tenant_port.assert_not_called()

    def test_expose_subnet(self):
        self.sb_idl.is_router_gateway_on_chassis.return_value = self.cr_lrp0
        self.sb_idl.get_port_datapath.return_value = 'fake-port-dp'
        row = fakes.create_object({
            'name': 'fake-row',
            'logical_port': 'subnet_port',
            'datapath': 'fake-dp',
            'options': {'peer': 'fake-peer'}})

        mock_expose_lrp_port = mock.patch.object(
            self.bgp_driver, '_expose_lrp_port').start()

        self.bgp_driver.expose_subnet('fake-ip', row)

        mock_expose_lrp_port.assert_called_once_with(
            'fake-ip', row.logical_port, self.cr_lrp0, 'fake-port-dp')

    def test_expose_subnet_no_cr_lrp(self):
        self.sb_idl.is_router_gateway_on_chassis.return_value = None
        self.sb_idl.get_port_datapath.return_value = 'fake-port-dp'
        row = fakes.create_object({
            'name': 'fake-row',
            'logical_port': 'subnet_port',
            'datapath': 'fake-dp',
            'options': {'peer': 'fake-peer'}})

        mock_expose_lrp_port = mock.patch.object(
            self.bgp_driver, '_expose_lrp_port').start()

        self.bgp_driver.expose_subnet('fake-ip', row)

        mock_expose_lrp_port.assert_not_called()

    def test_expose_subnet_address_scope(self):
        self.sb_idl.is_router_gateway_on_chassis.return_value = self.cr_lrp0
        self.sb_idl.get_port_datapath.return_value = 'fake-port-dp'
        row = fakes.create_object({
            'name': 'fake-row',
            'logical_port': 'subnet_port',
            'datapath': 'fake-dp',
            'options': {'peer': 'fake-peer'}})

        mock_expose_lrp_port = mock.patch.object(
            self.bgp_driver, '_expose_lrp_port').start()

        mock_address_scope_allowed = mock.patch.object(
            self.bgp_driver, '_address_scope_allowed').start()
        mock_address_scope_allowed.return_value = False

        self.bgp_driver.expose_subnet('fake-ip', row)

        mock_expose_lrp_port.assert_not_called()

    def test_withdraw_subnet(self):
        row = fakes.create_object({
            'name': 'fake-row',
            'logical_port': 'subnet_port',
            'datapath': 'fake-dp'})
        self.sb_idl.is_router_gateway_on_chassis.return_value = self.cr_lrp0
        mock_withdraw_lrp_port = mock.patch.object(
            self.bgp_driver, '_withdraw_lrp_port').start()

        self.bgp_driver.withdraw_subnet('{}/32'.format(self.ipv4), row)

        mock_withdraw_lrp_port.assert_called_once_with(
            '{}/32'.format(self.ipv4), row.logical_port, self.cr_lrp0)

    def test_withdraw_subnet_no_cr_lrp(self):
        row = fakes.create_object({
            'name': 'fake-row',
            'logical_port': 'subnet_port',
            'datapath': 'fake-dp'})
        self.sb_idl.is_router_gateway_on_chassis.return_value = None
        mock_withdraw_lrp_port = mock.patch.object(
            self.bgp_driver, '_withdraw_lrp_port').start()

        self.bgp_driver.withdraw_subnet('{}/32'.format(self.ipv4), row)

        mock_withdraw_lrp_port.assert_not_called()

    def test_withdraw_subnet_no_datapath_error(self):
        row = fakes.create_object({
            'name': 'fake-row',
            'logical_port': 'fake-logical-port',  # to match the cr-lrp name
            'datapath': 'fake-dp'})
        self.sb_idl.is_router_gateway_on_chassis.side_effect = (
            agent_exc.DatapathNotFound(datapath="fake-dp"))
        mock_withdraw_lrp_port = mock.patch.object(
            self.bgp_driver, '_withdraw_lrp_port').start()

        self.bgp_driver.withdraw_subnet('{}/32'.format(self.ipv4), row)

        mock_withdraw_lrp_port.assert_not_called()

    @mock.patch.object(linux_net, 'get_exposed_ips_on_network')
    @mock.patch.object(linux_net, 'delete_exposed_ips')
    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(linux_net, 'get_ip_version')
    def test__withdraw_lrp_port(
            self, mock_ip_version, mock_del_rule, mock_del_route,
            mock_del_exposed_ips, mock_get_exposed_ips):
        mock_ip_version.return_value = constants.IP_VERSION_4
        mock_get_exposed_ips.return_value = [self.ipv4]
        self.bgp_driver.ovn_local_lrps = {self.lrp0: self.cr_lrp0}

        self.bgp_driver._withdraw_lrp_port(
            '{}/32'.format(self.ipv4), self.lrp0, self.cr_lrp0)

        mock_del_rule.assert_called_once_with(
            '{}/32'.format(self.ipv4), 'fake-table')
        mock_del_route.assert_called_once_with(
            mock.ANY, self.ipv4, 'fake-table', self.bridge, vlan=None,
            mask='32', via=self.fip)
        mock_del_exposed_ips.assert_called_once_with(
            [self.ipv4], CONF.bgp_nic)

    @mock.patch.object(linux_net, 'get_exposed_ips_on_network')
    @mock.patch.object(linux_net, 'delete_exposed_ips')
    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(driver_utils, 'is_ipv6_gua')
    def test__withdraw_lrp_port_gua(
            self, mock_ipv6_gua, mock_ip_version, mock_del_rule,
            mock_del_route, mock_del_exposed_ips, mock_get_exposed_ips):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        CONF.set_override('expose_ipv6_gua_tenant_networks', True)
        self.addCleanup(CONF.clear_override, 'expose_ipv6_gua_tenant_networks')
        mock_ipv6_gua.return_value = True
        mock_ip_version.return_value = constants.IP_VERSION_6
        mock_get_exposed_ips.return_value = [self.ipv6]
        self.bgp_driver.ovn_local_lrps = {self.lrp0: self.cr_lrp0}

        self.bgp_driver._withdraw_lrp_port(
            '{}/128'.format(self.ipv6), self.lrp0, self.cr_lrp0)

        mock_del_rule.assert_called_once_with(
            '{}/128'.format(self.ipv6), 'fake-table')
        mock_del_route.assert_called_once_with(
            mock.ANY, self.ipv6, 'fake-table', self.bridge, vlan=None,
            mask='128', via=self.fip)
        mock_del_exposed_ips.assert_called_once_with(
            [self.ipv6], CONF.bgp_nic)

    @mock.patch.object(linux_net, 'get_exposed_ips_on_network')
    @mock.patch.object(linux_net, 'delete_exposed_ips')
    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'del_ip_rule')
    @mock.patch.object(driver_utils, 'is_ipv6_gua')
    def test__withdraw_lrp_port_no_gua(
            self, mock_ipv6_gua, mock_del_rule, mock_del_route,
            mock_del_exposed_ips, mock_get_exposed_ips):
        CONF.set_override('expose_tenant_networks', False)
        self.addCleanup(CONF.clear_override, 'expose_tenant_networks')
        CONF.set_override('expose_ipv6_gua_tenant_networks', True)
        self.addCleanup(CONF.clear_override, 'expose_ipv6_gua_tenant_networks')
        mock_ipv6_gua.return_value = False
        mock_get_exposed_ips.return_value = [self.ipv6]
        self.bgp_driver.ovn_local_lrps = {self.lrp0: self.cr_lrp0}

        self.bgp_driver._withdraw_lrp_port(
            'fdab:4ad8:e8fb:0:f816:3eff:fec6:469c/128', self.lrp0,
            self.cr_lrp0)

        mock_del_rule.assert_not_called()
        mock_del_route.assert_not_called()
        mock_del_exposed_ips.assert_not_called()

    @mock.patch.object(driver_utils, 'get_addr_scopes')
    def test__address_scope_allowed(self, m_addr_scopes):
        self.bgp_driver.allowed_address_scopes = {"fake_address_scope"}
        port_ip = self.ipv4
        port_name = "fake-port"
        sb_port = "fake-sb-port"
        self.sb_idl.get_port_by_name.return_value = sb_port
        address_scopes = {
            constants.IP_VERSION_4: "fake_address_scope",
            constants.IP_VERSION_6: "fake_ipv6_address_scope"}
        m_addr_scopes.return_value = address_scopes

        ret = self.bgp_driver._address_scope_allowed(port_ip, port_name)

        self.assertEqual(True, ret)
        m_addr_scopes.assert_called_once_with(sb_port)

    def test__address_scope_allowed_not_configured(self):
        self.bgp_driver.allowed_address_scopes = set()
        port_ip = self.ipv4
        port_name = "fake-port"
        sb_port = "fake-sb-port"

        ret = self.bgp_driver._address_scope_allowed(
            port_ip, port_name, sb_port)

        self.assertEqual(True, ret)

    @mock.patch.object(driver_utils, 'get_addr_scopes')
    def test__address_scope_allowed_no_match(self, m_addr_scopes):
        self.bgp_driver.allowed_address_scopes = {"fake_address_scope"}
        port_ip = self.ipv4
        port_name = "fake-port"
        sb_port = "fake-sb-port"
        self.sb_idl.get_port_by_name.return_value = sb_port
        address_scopes = {
            constants.IP_VERSION_4: "different_fake_address_scope",
            constants.IP_VERSION_6: "fake_ipv6_address_scope"}
        m_addr_scopes.return_value = address_scopes

        ret = self.bgp_driver._address_scope_allowed(port_ip, port_name)

        self.assertEqual(False, ret)
        m_addr_scopes.assert_called_once_with(sb_port)

    @mock.patch.object(driver_utils, 'get_addr_scopes')
    def test__address_scope_allowed_no_port(self, m_addr_scopes):
        self.bgp_driver.allowed_address_scopes = {"fake_address_scope"}
        port_ip = self.ipv4
        port_name = "fake-port"
        self.sb_idl.get_port_by_name.return_value = []

        ret = self.bgp_driver._address_scope_allowed(port_ip, port_name)

        self.assertEqual(False, ret)
        m_addr_scopes.assert_not_called()

    @mock.patch.object(driver_utils, 'get_addr_scopes')
    def test__address_scope_allowed_no_address_scope(self, m_addr_scopes):
        self.bgp_driver.allowed_address_scopes = {"fake_address_scope"}
        port_ip = self.ipv4
        port_name = "fake-port"
        sb_port = "fake-sb-port"
        self.sb_idl.get_port_by_name.return_value = sb_port
        address_scopes = {
            constants.IP_VERSION_4: "",
            constants.IP_VERSION_6: ""}
        m_addr_scopes.return_value = address_scopes

        ret = self.bgp_driver._address_scope_allowed(port_ip, port_name)

        self.assertEqual(False, ret)
        m_addr_scopes.assert_called_once_with(sb_port)
